/**
 * Copyright (C) 2003-2006 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the Honest Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Honest Public License for more details.
 *
 * You should have received a copy of the Honest Public License
 * along with this program; if not, write to Funambol,
 * 643 Bair Island Road, Suite 305 - Redwood City, CA 94063, USA
 */

package com.funambol.framework.engine;

import java.util.Map;
import java.util.HashMap;
import java.util.Iterator;

import com.funambol.framework.engine.source.SyncSource;
import com.funambol.framework.engine.SyncItem;
import com.funambol.framework.engine.SyncItemKey;
import com.funambol.framework.engine.SyncItemState;

import org.apache.commons.lang.builder.ToStringBuilder;
import java.sql.Timestamp;


/**
 * <i>SyncItem</i> is the indivisible entity that can be exchanged in a
 * synchronization process. It is similar to an Item, but it doesn't contain
 * any data, only status and addressing information. The idea is that a
 * <i>SyncItem</i> represents status information about an item. Only if an item
 * must be synchronized it needs also the real data.
 * <p>
 * The <i>SyncItemKey</i> uniquely identifies the item into the server. Client
 * keys must be translated into server keys before create a <i>SyncItem</i>.
 *
 *
 *
 * @version $Id: SyncItemImpl.java,v 1.7 2006/08/07 21:09:29 nichele Exp $
 *
 */
public class SyncItemImpl implements java.io.Serializable, SyncItem {

    // --------------------------------------------------------------- Constants
    public static final String PROPERTY_DECLARED_SIZE = "DECLARED_SIZE";

    public static final String PROPERTY_CONTENT_SIZE  = "CONTENT_SIZE";

    public static final String PROPERTY_COMMAND       = "PROPERTY_COMMAND";

    // -------------------------------------------------------------- Properties

    /**
     * The SyncItem's unique identifier - read only
     */
    protected SyncItemKey key = null;
    public SyncItemKey getKey() {
        return this.key;
    }

    public void setKey(SyncItemKey newKey) {
        this.key = newKey;
    }

    /**
     * The SyncItem's parent identifier - read only
     */
    protected SyncItemKey parentKey = null;
    public SyncItemKey getParentKey() {
        return this.parentKey;
    }

    public void setParentKey(SyncItemKey newParentKey) {
        this.parentKey = newParentKey;
    }


    /**
     * The SyncItem's mapped identifier - read/write
     */
    protected SyncItemKey mappedKey = null;
    public SyncItemKey getMappedKey() {
        return mappedKey;
    }

    /**
     *  A <i>SyncItem</i> is mapped if mappedKey is not null.
     *
     * @return <i>true</i> if the item is mapped to another source's item,
     *         <i>false</i> otherwise
     *
     */
    public boolean isMapped() {
        return (mappedKey != null);
    }

    /**
     * The state of this <i>SyncItem</i>
     *
     * @see com.funambol.framework.engine.SyncItemState
     */
    protected char state = SyncItemState.UNKNOWN;

    public char getState(){
        return state;
    }

    public void setState(char state){
        this.state = state;
    }

    /**
     * The content of this <i>SyncItem</i>
     *
     */
    protected byte[] content = null;

    public byte[] getContent(){
        return content;
    }

    public void setContent(byte[] content){
        this.content = content;
    }

    /**
     * The format of this <i>SyncItem</i>
     *
     */
    protected String format = null;

    public String getFormat(){
        return format;
    }

    public void setFormat(String format){
        this.format = format;
    }

    /**
     * The type of this <i>SyncItem</i>
     *
     */
    protected String type = null;

    public String getType(){
        return type;
    }

    public void setType(String type){
        this.type = type;
    }

    /**
     * The timestamp of this <i>SyncItem</i>
     *
     */
    protected Timestamp timestamp = null;

    public Timestamp getTimestamp(){
        return timestamp;
    }

    public void setTimestamp(Timestamp timestamp){
        this.timestamp = timestamp;
    }

    /**
     * The <i>SyncItem</i>'s properties - read and write
     */
    protected HashMap properties = new HashMap();

    /**
     * Returns the <i>properties</i> property. A cloned copy of the internal map
     * is returned.
     *
     * @return the <i>properties</i> property.
     */
    public Map getProperties() {
        return (Map)this.properties.clone();
    }

    /**
     * Sets the <i>properties</i> property. All items in the given map are added
     * to the internal map. <i>propertis</i> can contain String values or
     * <i>SyncPorperty</i> values. In the former case, <i>new SyncProperty<i/>s are
     * created.
     *
     * @param properties the new values
     */
    public void setProperties(Map properties){
        this.properties.clear();

        Object value = null;
        String name  = null;

        Iterator i = properties.keySet().iterator();
        while(i.hasNext()) {
            name  = (String)i.next();
            value = properties.get(name);

            if (!(value instanceof SyncProperty)) {
                value = new SyncProperty(name, value.toString());
            }
            this.properties.put(name, value);
        }
    }

    /** Sets/adds the given property to this <i>SyncItem</i>
     *
     * @param property The property to set/add
     */
    public void setProperty(SyncProperty property) {
        properties.put(property.getName(), property);
    }

    /** Returns the property with the given name
     *
     * @param propertyName The property name
     *
     * @return the property with the given name if exists or null if not
     */
    public SyncProperty getProperty(String propertyName) {
        return (SyncProperty)properties.get(propertyName);
    }

        /**
     * The SyncSource this item belongs to
     */
    protected SyncSource syncSource = null;

    /** Getter for property syncSource.
     * @return Value of property syncSource.
     *
     */
    public SyncSource getSyncSource() {
        return syncSource;
    }

    /** Setter for property syncSource.
     * @param syncSource New value of property syncSource. NOT NULL
     *
     */
    public void setSyncSource(SyncSource syncSource) {
        if (syncSource == null) {
            throw new NullPointerException("syncSource cannot be null");
        }

        this.syncSource = syncSource;
    }

    // ------------------------------------------------------------ Constructors

    protected SyncItemImpl() {}

    /**
     * Creates a new <i>SyncItem</i> belonging to the given source with the given
     * key.
     *
     * @param syncSource the source this item belongs to
     * @param key the item identifier
     */
    public SyncItemImpl(SyncSource syncSource, Object key) {
        this(syncSource,
             key,
             null,  // parentKey
             null,  // mappedKey
             SyncItemState.UNKNOWN,
             null,  // content
             null,  // format
             null,  // type
             null   // timestamp
             );
    }

    /**
     * Creates a new <i>SyncItem</i> belonging to the given source with the given
     * key and the given state.
     *
     * @param syncSource the source this item belongs to
     * @param key the item identifier
     * @param state one of the state value defined in <i>SyncItemState</i>
     */
    public SyncItemImpl(SyncSource syncSource, Object key, char state) {
        this(syncSource,
             key,
             null,  // parentKey
             null,  // mappedKey
             state,
             null,  // content
             null,  // format
             null,  // type
             null   // timestamp
             );
    }

    /**
     * Creates a new <i>SyncItem</i> belonging to the given source and with the given
     * key, parentKey and state.
     *
     * @param syncSource the source this item belongs to
     * @param key the item identifier
     * @param parentKey the item's parentKey
     * @param state one of the state value defined in <i>SyncItemState</i>
     */
    public SyncItemImpl(SyncSource syncSource, Object key, Object parentKey, char state) {
        this(syncSource,
             key,
             parentKey,
             null,  // mappedKey
             state,
             null,  // content
             null,  // format
             null,  // type
             null   // timestamp
             );
    }


    /**
     * Creates a new <i>SyncItem</i> belonging to the given source with the
     * given key, parentKey, mappedKey and state.
     *
     * @param syncSource the source this item belongs to
     * @param key the item identifier
     * @param parentKey the parent key of the item
     * @param mappedKey the mapped key
     * @param state one of the state value defined in <i>SyncItemState</i>
     */
    public SyncItemImpl(SyncSource syncSource,
                        Object     key,
                        Object     parentKey,
                        Object     mappedKey,
                        char       state) {

        this(syncSource,
             key,
             parentKey,
             mappedKey,
             state,
             null, // content
             null, // format
             null, // type
             null  // timestamp
        );

    }

    /**
     * Creates a new <i>SyncItem</i>.
     *
     * @param syncSource the source this item belongs to
     * @param key the item identifier
     * @param mappedKey the mapped key
     * @param state one of the state value defined in <i>SyncItemState</i>
     * @param content the content of the item
     * @param format the format of the item
     * @param type the type of the item
     * @param timestamp the timestamp of the item
     */
    public SyncItemImpl(SyncSource syncSource,
                        Object     key,
                        Object     mappedKey,
                        char       state,
                        byte[]     content,
                        String     format,
                        String     type,
                        Timestamp  timestamp) {

        this(syncSource,
             key,
             null, // parentKey
             mappedKey,
             state,
             content,
             format,
             type,
             timestamp
        );
    }

    /**
     * Creates a new <i>SyncItem</i>.
     *
     * @param syncSource the source this item belongs to
     * @param key the item identifier
     * @param parentKey the item's parentKey
     * @param mappedKey the mapped key
     * @param state one of the state value defined in <i>SyncItemState</i>
     * @param content the content of the item
     * @param format the format of the item
     * @param type the type of the item
     * @param timestamp the timestamp of the item
     */
    public SyncItemImpl(SyncSource syncSource,
                        Object     key,
                        Object     parentKey,
                        Object     mappedKey,
                        char       state,
                        byte[]     content,
                        String     format,
                        String     type,
                        Timestamp  timestamp) {

        this.syncSource = syncSource          ;
        this.key        = new SyncItemKey(key);
        this.state      = state               ;

        this.parentKey = (parentKey != null)
                         ? new SyncItemKey(parentKey)
                         : null;

        this.mappedKey = (mappedKey != null)
                         ? new SyncItemKey(mappedKey)
                         : null;

        this.content   = content;
        this.format    = format;
        this.type      = type;
        this.timestamp = timestamp;
    }

    // ---------------------------------------------------------- Static Methods
    /**
     * Returns a clone of the given item
     * @param item the item to clone
     * @return the clone of the given item
     */
    public static SyncItemImpl cloneSyncItem(SyncItemImpl item) {

        if (item == null) {
            return null;
        }

        String key = (item.getKey() != null) ? item.getKey().getKeyAsString() : null;

        String parentKey =
            (item.getParentKey() != null) ? item.getParentKey().getKeyAsString() : null;

        String mappedKey =
            (item.getMappedKey() != null) ? item.getMappedKey().getKeyAsString() : null;

        SyncItemImpl clone = new SyncItemImpl(item.getSyncSource(),
                                              key,
                                              parentKey,
                                              mappedKey,
                                              item.getState(),
                                              item.getContent(),
                                              item.getFormat(),
                                              item.getType(),
                                              item.getTimestamp());

        clone.setProperties(item.getProperties());
        return clone;
    }

    /**
     * Creates and returns a "not-existing" <i>SyncItem</i>. It is used internally to
     * represent an item which has not a physical correspondance in a source.
     *
     * @param syncSource the <i>SyncSource</i> the not existing item belongs to
     * @return the a "not-exisiting" <i>SyncItem</i>
     */
    public static SyncItem getNotExistingSyncItem(SyncSource syncSource) {
        SyncItem notExisting = new SyncItemImpl(syncSource, "NOT_EXISTING");

        notExisting.setState(SyncItemState.NOT_EXISTING);

        return notExisting;
    }

    // ---------------------------------------------------------- Public methods


    /** Sets the value of the property with the given name.
     *
     * @param propertyName The property's name
     * @param propertyValue The new value
     */
    public void setPropertyValue(String propertyName, Object propertyValue) {
        SyncProperty property = (SyncProperty)properties.get(propertyName);

        if (property != null) {
            property.setValue(propertyValue);
        }
    }

    /** Returns the value of the property with the given name.
     *
     * @param propertyName The property's name
     *
     * @return the property value if this <i>SyncItem</i> has the given
     *         property or null otherwise.
     */
    public Object getPropertyValue(String propertyName) {
        SyncProperty property = (SyncProperty)properties.get(propertyName);

        return (property == null) ? null
                                  : property.getValue()
                                  ;
    }

    /**
     * Two <i>SyncItem</i>s are equal if their keys are equal.
     *
     * @param o the object this instance must be compared to.
     *
     * @return the given object is equal to this object
     *
     */
    public boolean equals(Object o) {
        if (!(o instanceof SyncItem)) return false;

        return ((SyncItem)o).getKey().equals(key);
    }

    /**
     * Returns the hash code of this object
     *
     * @return the hash code of this object
     */
    public int hashCode() {
        return getKey().hashCode();
    }

    /**
     * Returns a string representation of this SyncItem
     * @return a string representation of this SyncItem for debugging purposes
     */
    public String toString() {
        return new ToStringBuilder(this).
                   append("key"       ,
                          (key != null ? key.toString() : "null" )).

                   append("parentKey" ,
                          (parentKey != null ? parentKey.toString() : "null" )).

                   append("mappedKey" ,
                          (mappedKey != null ? mappedKey.toString() : "null" )).

                   append("format"    , format                ).
                   append("type"      , type                  ).
                   append("timestamp" , timestamp             ).

                   append("content-lenght",
                          (content != null ? String.valueOf(content.length) : "content null")).

                   append("state"     , state                 ).
                   append("properties",  properties.toString()).
                   toString();
    }


}
