/**
 * Copyright (C) 2003-2006 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the Honest Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Honest Public License for more details.
 *
 * You should have received a copy of the Honest Public License
 * along with this program; if not, write to Funambol,
 * 643 Bair Island Road, Suite 305 - Redwood City, CA 94063, USA
 */

package com.funambol.framework.engine.source;

import java.sql.Timestamp;

import java.util.logging.Level;
import java.util.logging.Logger;

import com.funambol.framework.engine.SyncItem;
import com.funambol.framework.engine.SyncItemKey;
import com.funambol.framework.logging.Sync4jLogger;

/**
 * Implements the basic functionalities of a <i>SyncSource</i> like naming.
 *
 * @version $Id: AbstractSyncSource.java,v 1.5 2006/08/07 21:09:30 nichele Exp $
 */
public abstract class AbstractSyncSource implements SyncSource, java.io.Serializable {

    // --------------------------------------------------------------- Constants
    public static final String LOG_NAME = "engine.source";

    // ---------------------------------------------------------- Protected data

    protected String name         = null;
    protected String type         = null;
    protected String sourceURI    = null;
    protected String sourceQuery  = null;
    protected SyncSourceInfo info = null;

    /**
     * How many items were added ?
     */
    protected long howManyAdded;

    /**
     * How many items were deleted ?
     */
    protected long howManyDeleted;

    /**
     * How many items were updated ?
     */
    protected long howManyUpdated;

    // ------------------------------------------------------------ Constructors

    /** Creates a new instance of AbstractSyncSource */
    protected AbstractSyncSource() {
    }

    public AbstractSyncSource(String name, String type, String sourceURI) {
        this.name      = name;
        this.type      = (type == null) ? "unknown" : type;
        this.sourceURI = sourceURI;
    }

    public AbstractSyncSource(String name) {
        this(name, null, null);
    }

    // ---------------------------------------------------------- Public methods

    public String getName() {
        return name;
    }

    public void setName(String name) {
        this.name = name;
    }

    public String getType() {
        return this.type;
    }

    public void setType(String type) {
        this.type = type;
    }

    /**
     * Getter for property query.
     * @return Value of property query.
     */
    public String getSourceQuery() {
        return sourceQuery;
    }

    /**
     * Getter for property uri.
     * @return Value of property uri.
     */
    public String getSourceURI() {
        return sourceURI;
    }

    /**
     * Setter for property uri.
     * @param sourceURI New value of property uri.
     */
    public void setSourceURI(String sourceURI) {
        this.sourceURI = sourceURI;
    }

    /**
     * Returns the type info of the content handled by this source
     *
     * @return the type info of the content handled by this source
     */
    public SyncSourceInfo getInfo() {
        return this.info;
    }

    /**
     * Setter for the property <i>info</i>
     */
    public void setInfo(SyncSourceInfo info) {
        this.info = info;
    }

    /**
     * Returns a string representation of this object.
     *
     * @return a string representation of this object.
     */
    public String toString() {
        StringBuffer sb = new StringBuffer(super.toString());

        sb.append(" - {name: ").append(getName()     );
        sb.append(", type: "   ).append(getType()     );
        sb.append(", uri: "    ).append(getSourceURI());
        sb.append("}"         );

        return sb.toString();
    }

    /**
     * This resets the howMany counters. It should then be called by extending
     * classes before anything else.
     *
     * @param syncContext the syncContext to use
     *
     * @throws SyncSourceException in case of any error
     *
     * @see SyncSource
     */
    public void beginSync(SyncContext syncContext) throws SyncSourceException {

        howManyAdded = howManyDeleted = howManyUpdated = 0;
    }

    /**
     * It logs the howMany counters. It should then be called by extending
     * classes before anything else.
     *
     * @param principal user/device
     *
     * @throws SyncSourceException in case of any error
     *
     * @see SyncSource
     */
    public void endSync() throws SyncSourceException {
        Logger log = Sync4jLogger.getLogger(LOG_NAME);

        if (log.isLoggable(Level.INFO)) {
            log.info(getSourceURI()
                    + ": "
                    + howManyAdded
                    + " new items added, "
                    + howManyUpdated
                    + " existing items updated, "
                    + howManyDeleted
                    + " items deleted."
                    );
        }
    }

    /**
     * @see SyncSource
     */
    public void commitSync() throws SyncSourceException {
    }

    // -------------------------------------------------------- Abstract methods

    public abstract SyncItemKey[] getAllSyncItemKeys() throws SyncSourceException;

    public abstract SyncItemKey[] getDeletedSyncItemKeys(Timestamp sinceTs,
                                                         Timestamp untilTs)
        throws SyncSourceException;

    public abstract SyncItemKey[] getNewSyncItemKeys(Timestamp sinceTs,
                                                     Timestamp untilTs)
        throws SyncSourceException;


    public abstract SyncItem getSyncItemFromId(SyncItemKey syncItemKey)
        throws SyncSourceException;

    public abstract void removeSyncItem(SyncItemKey syncItemKey,
                                        Timestamp   time,
                                        boolean     softDelete) throws SyncSourceException;

    public abstract SyncItem updateSyncItem(SyncItem syncInstance)
        throws SyncSourceException;

    public abstract SyncItem addSyncItem(SyncItem syncInstance)
        throws SyncSourceException;

    public abstract SyncItemKey[] getSyncItemKeysFromTwin(SyncItem syncItem)
        throws SyncSourceException;

    public abstract void setOperationStatus(String operationName,
                                            int status,
                                            SyncItemKey[] keys);

}
