/**
 * Copyright (C) 2006 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the Honest Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Honest Public License for more details.
 *
 * You should have received a copy of the Honest Public License
 * along with this program; if not, write to Funambol,
 * 643 Bair Island Road, Suite 305 - Redwood City, CA 94063, USA
 */
package com.funambol.framework.engine.transformer;

import java.util.HashMap;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import com.funambol.framework.engine.SyncItem;
import com.funambol.framework.logging.Sync4jLogger;

/**
 * This class represents a data encoder or decoder
 *
 * A DataTransformer is a component that given some data in some format
 * as input, it generates a transformation of the source data in output.
 * Examples of transformations are:
 *
 * - base64 encoding/decoding
 * - encryption/decription
 */
public class DataTransformerManager {

    // ------------------------------------------------------------ Private data
    private static final Logger log = Sync4jLogger.getLogger("engine");

    /**
     * Contains the source uri and the relative transformations required
     */
    private Map sourceUriTrasformationsRequired;

    /**
     * Contains the transformer configured for the incoming items
     */
    private Map dataTransformersIn;

    /**
     * Contains the transformer configured for the outgoing items
     */
    private Map dataTransformersOut;

    // ------------------------------------------------------------ Constructors


    /** Creates a new instance of DataTransformerManager */
    public DataTransformerManager() {
        sourceUriTrasformationsRequired = new HashMap();
        dataTransformersIn  = new HashMap();
        dataTransformersOut = new HashMap();
    }

    // ---------------------------------------------------------- Public methods

    public Map getSourceUriTrasformationsRequired() {
        return sourceUriTrasformationsRequired;
    }

    public void setSourceUriTrasformationsRequired(Map sourceUriTrasformationsRequired) {
        this.sourceUriTrasformationsRequired = sourceUriTrasformationsRequired;
    }

    public Map getDataTransformersIn() {
        return dataTransformersIn;
    }


    public void setDataTransformersIn(Map dataTransformersIn) {
        this.dataTransformersIn = dataTransformersIn;
    }


    public Map getDataTransformersOut() {
        return dataTransformersOut;
    }


    public void setDataTransformersOut(Map dataTransformersOut) {
        this.dataTransformersOut = dataTransformersOut;
    }


    /**
     * Applies on the given item the transformations required based on his source uri
     * using the data contained in the given info.
     * @param info the transformation info to use
     * @param item the item to convert
     * @return the item converted
     * @throws TransformerException if some errors occur
     */
    public SyncItem transformIngoingItem(TransformationInfo info, SyncItem item)
    throws TransformerException {

        String format = item.getFormat();
        if (format == null) {
            //
            // No transformations are applicable
            //
            return item;
        }

        String uri    = item.getSyncSource().getSourceURI();

        String transformerNames = (String)sourceUriTrasformationsRequired.get(uri);
        if (transformerNames == null) {
            //
            // no transformations are required
            //
            return item;
        }

        if (format.endsWith(transformerNames)) {
            //
            // We have to apply the transformes specifies in the transformerNames
            // in inverse order
            //
            if (log.isLoggable(Level.FINEST)) {
                log.finest("Applying transformations '" + transformerNames + "'");
            }

            // textplain;des;b64
            byte[]          content  = item.getContent();
            String[]        tokens   = transformerNames.split(";");
            int             numToken = tokens.length;
            DataTransformer dt       = null;
            for (int i=numToken-1; i>=0; i--) {

                if (log.isLoggable(Level.FINEST)) {
                    log.finest("Applying transformation: '" + tokens[i] + "'");
                }

                dt = getDataTransformerIn(tokens[i]);

                if (dt != null) {
                    content = dt.transform(content, info);
                    if (log.isLoggable(Level.FINEST)) {
                        log.finest("Transformation '" + tokens[i] + "' done");
                    }
                } else {
                    if (log.isLoggable(Level.FINEST)) {
                        log.finest("DataTransformer for '" + tokens[i] +
                                   "' for item goes in not found");
                    }
                }
            }
            item.setContent(content);

            format = format.substring(0, format.indexOf(transformerNames));
            item.setFormat(format);
        }


        return item;
    }

    /**
     * Applies on the given item the transformations required based on his source uri
     * using the data contained in the given info.
     * @param info the transformation info to use
     * @param item the item to convert
     * @return the item converted
     * @throws TransformerException if some errors occur
     */
    public SyncItem transformOutgoingItem(TransformationInfo info, SyncItem item)
    throws TransformerException {

        String format = item.getFormat();
        String uri    = item.getSyncSource().getSourceURI();

        String transformerNames = (String)sourceUriTrasformationsRequired.get(uri);
        if (transformerNames == null) {
            //
            // no transformations are required
            //
            return item;
        }

        //
        // We have to apply the transformes specifies in the transformerNames
        // in inverse order
        //
        if (log.isLoggable(Level.FINEST)) {
            log.finest("Applying transformations '" + transformerNames + "'");
        }

        // textplain;des;b64
        if (format == null) {
            format = "";
        }
        StringBuffer    sb       = new StringBuffer(format);
        byte[]          content  = item.getContent();
        String[]        tokens   = transformerNames.split(";");
        int             numToken = tokens.length;
        DataTransformer dt       = null;
        for (int i=0; i<numToken; i++) {

            if (log.isLoggable(Level.FINEST)) {
                log.finest("Applying transformation: '" + tokens[i] + "'");
            }

            dt = getDataTransformerOut(tokens[i]);

            if (dt != null) {
                content = dt.transform(content, info);
                if (sb.length() != 0) {
                    sb.append(";");
                }
                sb.append(tokens[i]);
                if (log.isLoggable(Level.FINEST)) {
                    log.finest("Transformation '" + tokens[i] + "' done");
                }
            } else {
                if (log.isLoggable(Level.FINEST)) {
                    log.finest("DataTransformer for '" + tokens[i] +
                               "' for item goes in not found");
                }
                throw new TransformerException("DataTransformer for '" + tokens[i] +
                                               "' for item goes in not found");
            }
        }
        item.setContent(content);
        item.setFormat(sb.toString());

        return item;
    }

    // --------------------------------------------------------- Private Methods

    /**
     * Returns the DataTransformer given the name for incoming items
     * @param name the transformer name
     * @return the data transformer
     * @throws if some errors occur
     */
    private DataTransformer getDataTransformerIn(String name)
    throws TransformerException {
        String          dtClass = (String)dataTransformersIn.get(name);
        DataTransformer dt      = null;

        if (log.isLoggable(Level.FINEST)) {
            log.finest("DataTransformer class for '" + name + "' in input: " + dtClass);
        }
        if (dtClass != null) {
            try {
                dt = (DataTransformer) Class.forName(dtClass).newInstance();
            } catch (Exception ex) {
                log.throwing(this.getClass().getName(), "getDataTransformerIn", ex);
                throw new TransformerException("Unable to create an instance of '" + dtClass + "'");
            }
        }
        if (log.isLoggable(Level.FINEST)) {
            log.finest("DataTransformer instance for '" + name + "' in input: " + dt);
        }
        return dt;
    }


    /**
     * Returns the DataTransformer given the name for outgoing items
     * @param name the transformer name
     * @return the data transformer
     * @throws if some errors occur
     */
    private DataTransformer getDataTransformerOut(String name)
    throws TransformerException {
        String          dtClass = (String)dataTransformersOut.get(name);
        DataTransformer dt      = null;

        if (log.isLoggable(Level.FINEST)) {
            log.finest("DataTransformer class for '" + name + "' in output: " + dtClass);
        }
        if (dtClass != null) {
            try {
                dt = (DataTransformer) Class.forName(dtClass).newInstance();
            } catch (Exception ex) {
                log.throwing(this.getClass().getName(), "getDataTransformerOut", ex);
                throw new TransformerException("Unable to create an instance of '" + dtClass + "'");
            }
        }
        if (log.isLoggable(Level.FINEST)) {
            log.finest("DataTransformer instance for '" + name + "' in output: " + dt);
        }
        return dt;
    }
}
