/**
 * Copyright (C) 2003-2006 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the Honest Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Honest Public License for more details.
 *
 * You should have received a copy of the Honest Public License
 * along with this program; if not, write to Funambol,
 * 643 Bair Island Road, Suite 305 - Redwood City, CA 94063, USA
 */

package com.funambol.framework.server.store;

import java.util.Map;
import java.util.logging.Logger;

import java.sql.*;

import javax.naming.InitialContext;
import javax.naming.NameNotFoundException;
import javax.naming.NamingException;
import javax.sql.DataSource;

import com.funambol.framework.server.store.PersistentStoreException;
import com.funambol.framework.logging.Sync4jLogger;
import com.funambol.framework.tools.DBTools;

import org.apache.commons.lang.StringUtils;

/**
 * This is a base class for <i>PersistenStore</i> objects. It does not persist
 * any object, but it provides services common to concrete implementations.
 *
 * @version $Id: BasePersistentStore.java,v 1.3 2006/09/06 16:40:27 nichele Exp $
 *
 */
public abstract class BasePersistentStore {

    // --------------------------------------------------------------- Constants

    public static final String CONFIG_JNDI_DATA_SOURCE_NAME = "jndi-data-source-name";

    /**
     * Logger
     */
    protected transient final Logger log = Sync4jLogger.getLogger();

    // -------------------------------------------------------------- Properties

    /**
     * The database product name
     */
    private String dbProductName = null;

    /**
     * The JNDI name of the datasource to be used
     */
    protected String jndiDataSourceName = null;

    public String getJndiDataSourceName() {
        return this.jndiDataSourceName;
    }

    public void setJndiDataSourceName(String jndiDataSourceName) throws PersistentStoreException {
        this.jndiDataSourceName = jndiDataSourceName;

        if (jndiDataSourceName == null) {
            dataSource = null;
        }

        //
        // we first try with the name as it is; if we get a NameNotFound
        // exception than we try prepending 'java:/comp/env/'
        //
        try {
            InitialContext ctx = new InitialContext();
            try {
                dataSource = (DataSource) ctx.lookup(jndiDataSourceName);
            } catch (NameNotFoundException e) {
                if (jndiDataSourceName.startsWith("java:/")) {
                    jndiDataSourceName = (jndiDataSourceName.length()>6)
                                       ? jndiDataSourceName.substring(6)
                                       : jndiDataSourceName
                                       ;
                }

                dataSource = (DataSource) ctx.lookup("java:/comp/env/" + jndiDataSourceName);
            }
        } catch (NamingException e) {
            throw new PersistentStoreException("Data source "
            + jndiDataSourceName
            + " not found"
            , e
            );
        }

        Connection conn = null;
        try {
            conn = dataSource.getConnection();
            dbProductName = conn.getMetaData().getDatabaseProductName();
            sqlHelperClause = new SQLHelperClause(dbProductName);
        } catch(SQLException e) {
            log.throwing("BasePersistentStore", "setJndiDataSourceName", e);
            throw new PersistentStoreException("Error getting connection (" +
                                               e.getMessage() + ")", e);
        } finally {
            DBTools.close(conn, null, null);
        }
    }

    // ---------------------------------------------------------- Protected data

    protected transient DataSource dataSource = null;
    protected SQLHelperClause sqlHelperClause = null;

    // ------------------------------------------------------------ Constructors

    // ---------------------------------------------------------- Public methods

    /** Configures the persistent store
     *
     * @param config an <i>Map</i> containing configuration parameters.
     *
     * @throws ConfigPersistentStoreException
     *
     */
    public void configure(Map config) throws ConfigPersistentStoreException {

        checkConfigParams(config);

        try {
            setJndiDataSourceName((String) config.get(CONFIG_JNDI_DATA_SOURCE_NAME));
        } catch (PersistentStoreException e) {
            throw new ConfigPersistentStoreException( "Error creating the datasource: "
                                                    + e.getMessage()
                                                    , e
                                                    );
        }
    }

    // ------------------------------------------------------- Protected methods

    // ------------------------------------------------------- Protected methods

    // --------------------------------------------------------- Private methods

    /**
     * Checks if the given configuration parameters contain all required
     * parameters. If not a <i>ConfigPersistentStoreException</i> is thrown.
     *
     * @param config the <i>Map</i> containing the configuration parameters
     *
     * @throws ConfigPersistentStoreException in case of missing parameters
     */
    private void checkConfigParams(Map config)
    throws ConfigPersistentStoreException {
        StringBuffer sb = new StringBuffer();

        if (StringUtils.isEmpty((String) config.get(CONFIG_JNDI_DATA_SOURCE_NAME))) {
            sb.append(CONFIG_JNDI_DATA_SOURCE_NAME);
        }
    }
}

