/**
 * Copyright (C) 2003-2006 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the Honest Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Honest Public License for more details.
 *
 * You should have received a copy of the Honest Public License
 * along with this program; if not, write to Funambol,
 * 643 Bair Island Road, Suite 305 - Redwood City, CA 94063, USA
 */
package com.funambol.server.config;

import java.io.IOException;
import java.io.File;
import java.io.StringWriter;
import java.io.PrintWriter;

import java.net.*;

import java.util.Iterator;
import java.util.logging.ConsoleHandler;
import java.util.logging.FileHandler;
import java.util.logging.Handler;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.logging.LogManager;

import com.funambol.framework.config.*;
import com.funambol.framework.engine.SyncStrategy;
import com.funambol.framework.engine.pipeline.PipelineManager;
import com.funambol.framework.engine.transformer.DataTransformerManager;
import com.funambol.framework.logging.Sync4jLogger;
import com.funambol.framework.security.Officer;
import com.funambol.framework.server.inventory.DeviceInventory;
import com.funambol.framework.server.session.SessionHandler;
import com.funambol.framework.server.store.PersistentStore;
import com.funambol.framework.tools.beans.BeanFactory;
import com.funambol.framework.tools.beans.BeanException;

import com.funambol.server.admin.UserManager;

/**
 * Incapsulates all the configuration information about the Sync4j Engine.
 * Configuration parameters are stored as properties in a Map structure. The
 * key is the name of the parameter and it is strucured in dotted separated
 * namespaces. The value of the parameter is the value of the key and can be of
 * any type. Accessor methods are provided to get the value of the parameter in
 * a particular type (such as String, int, double, boolean, ...).<br>
 * Access to getXXX() and setXXX() methods are synchronized.
 * <p>
 * This class follows the Singleton pattern, therefore it cannot be directly
 * instantiated; an internal instance is created just once and returned at any
 * getConfiguration() call.
 *
 * @version $Id: Configuration.java,v 1.4 2006/08/10 22:14:13 nichele Exp $
 */
public class Configuration
implements java.io.Serializable, ConfigurationConstants {

    // ------------------------------------------------------------ Private data

    /**
     * The server configuration
     */
    private static Configuration singleton;
    private transient ClassLoader classLoader = null;

    private static transient Logger log = Sync4jLogger.getLogger("configuration");

    // -------------------------------------------------------------- properties

    /** Getter for property classLoader.
     * @return Value of property classLoader.
     *
     */
    public ClassLoader getClassLoader() {
        return classLoader;
    }

    /** Setter for property classLoader.
     * @param classLoader New value of property classLoader.
     *
     */
    public void setClassLoader(ClassLoader classLoader) {
        this.classLoader = classLoader;
    }

    // ------------------------------------------------------------ Constructors

    /**
     * This is disabled for normal use since this class is a singleton.
     */
    protected Configuration() {
    }

    /**
     * Returns the singleton instance. It created the instance the first time it
     * is called.
     *
     * @return the singleton instance.
     */
    public static synchronized Configuration getConfiguration() {
        if (singleton == null) {
            singleton = new Configuration();
            try {
                URL configPath = ConfigTools.fixURL(getConfigPath());
                ConfigClassLoader cl = new ConfigClassLoader(
                                           new URL[] { configPath },
                                           Configuration.class.getClassLoader()

                                       );

                singleton.setClassLoader(cl);

                singleton.setLoggingConfiguration();

            } catch (Exception e) {
                String msg = "Fatal error creating the configuration object ";
                if (log.isLoggable(Level.SEVERE)) {
                    log.severe(msg + "(" + e.toString() + ")");
                }
                log.throwing(Configuration.class.getName(), "getConfiguration", e);

                //
                // In this way the exception is logged only once because if the log
                // level is FINER then the exception is logged by Throwing.
                // Knowing the stacktrace of this exception is very useful so
                // it is logged also if the level is INFO
                //
                if (log.isLoggable(Level.INFO) && !log.isLoggable(Level.FINER)) {
                    StringWriter sw = new StringWriter();
                    e.printStackTrace(new PrintWriter(sw));
                    log.info(sw.toString());
                }
            }
        }
        return singleton;
    }

    // ---------------------------------------------------------- Public methods

    /**
     * Getter for property serverConfig.
     * @return Value of property serverConfig.
     */
    public ServerConfiguration getServerConfig() {
        return (ServerConfiguration)ConfigTools.getBeanInstance(classLoader, BEAN_SERVER_CONFIGURATION);
    }

    /**
     * Load the server configuration
     */
    public void loadServerConfig() {
        //
        // In order to load the server configuration, we call the static
        // synchronized getConfiguration method after we have set the singleton object
        // to null. So, the getConfiguration method reload the server configuration bean
        // and re-set all the properties depends on it
        //
        singleton = null;
        getConfiguration();
    }

    /**
     * Creates and returns a new SyncStrategy instance from the configured
     * <i>singleton.getServerConfig().getEngineConfiguration().strategy</i> property
     *
     * @return the newly created instance
     */
    public SyncStrategy getStrategy() {
        try {
            return (SyncStrategy)ConfigTools.getNewBeanInstance(
                classLoader,
                singleton.getServerConfig().getEngineConfiguration().getStrategy()
            );
        } catch (Exception e) {
            if (log.isLoggable(Level.SEVERE)) {
                log.severe("Error creating the SyncStrategy object: " + e.getMessage());
            }
            log.throwing(getClass().getName(), "getStartegy", e);
        }

        return null;
    }

    /**
     * Creates and returns a new PersistentStore instance from the configured
     * <i>singleton.getServerConfig().getEngineConfiguration().store</i> property
     *
     * @return the newly created instance
     */
    public PersistentStore getStore() {
        try {
            return (PersistentStore)ConfigTools.getBeanInstance(
                classLoader,
                singleton.getServerConfig().getEngineConfiguration().getStore()
            );
        } catch (Exception e) {
            if (log.isLoggable(Level.SEVERE)) {
                log.severe("Error creating the PersistentStore object: " + e.getMessage());
            }
            log.throwing(getClass().getName(), "getStore", e);
        }

        return null;
    }

    /**
     * Creates and returns a new Officer instance from the configured
     * <i>singleton.getServerConfig().getEngineConfiguration().getOfficer</i> property
     *
     * @return the newly created instance
     */
    public Officer getOfficer() {
        try {
            return (Officer)ConfigTools.getBeanInstance(
                classLoader,
                singleton.getServerConfig().getEngineConfiguration().getOfficer()
            );
        } catch (Exception e) {
            if (log.isLoggable(Level.SEVERE)) {
                log.severe("Error creating the Officer object: " + e.getMessage());
            }
            log.throwing(getClass().getName(), "getOfficer", e);
        }

        return null;
    }


    /**
     * Creates and returns a new SessionHandler instance from the configured
     * <i>singleton.getServerConfig().getEngineConfiguration().sessionHandler</i> property
     *
     * @return the newly created instance
     */
    public SessionHandler getSessionHandler() {
        try {
            return (SessionHandler)ConfigTools.getNewBeanInstance(
                classLoader,
                singleton.getServerConfig().getEngineConfiguration().getSessionHandler()
            );
        } catch (Exception e) {
            if (log.isLoggable(Level.SEVERE)) {
                log.severe("Error creating the SessionHandler object: " + e.getMessage());
            }
            log.throwing(getClass().getName(), "getSessionHandler", e);
        }

        return null;
    }

    /**
     * Creates and returns a new PipelineManager instance from the configured
     * <i>singleton.getServerConfig().getEngineConfiguration().pipelineManager</i> property
     *
     * @return the newly created instance
     */
    public PipelineManager getPipelineManager() {
        try {
            return (PipelineManager)ConfigTools.getNewBeanInstance(
                classLoader,
                singleton.getServerConfig().getEngineConfiguration().getPipelineManager()
            );
        } catch (Exception e) {
            if (log.isLoggable(Level.SEVERE)) {
                log.severe("Error creating the PipelineManager object: " + e.getMessage());
            }
            log.throwing(getClass().getName(), "getPipelineManager", e);
        }

        return null;
    }

    /**
     * Creates and returns a new UserManager instance from the configured
     * <i>singleton.getServerConfig().getEngineConfiguration().userManager</i> property
     *
     * @return the newly created instance
     */
    public UserManager getUserManager() {
        try {
            return (UserManager)ConfigTools.getNewBeanInstance(
                classLoader,
                singleton.getServerConfig().getEngineConfiguration().getUserManager()
            );
        } catch (Exception e) {
            if (log.isLoggable(Level.SEVERE)) {
                log.severe("Error creating the UserManager object: " + e.getMessage());
            }
            log.throwing(getClass().getName(), "getUserManager", e);
        }

        return null;
    }


    /**
     * Creates and returns a new PipelineManager instance from the configured
     * <i>serverConfig.getEngineConfiguration().pipelineManager</i> property
     *
     * @return the newly created instance
     */
    public DeviceInventory getDeviceInventory() {
        try {
            return (DeviceInventory)ConfigTools.getBeanInstance(
                classLoader,
                singleton.getServerConfig().getEngineConfiguration().getDeviceInventory()
            );
        } catch (Exception e) {
            if (log.isLoggable(Level.SEVERE)) {
                log.severe("Error creating the DeviceInventory object: " + e.getMessage());
            }
            log.throwing(getClass().getName(), "getDeviceInventory", e);
        }

        return null;
    }

    /**
     * Creates and returns a new PipelineManager instance from the configured
     * <i>serverConfig.getEngineConfiguration().pipelineManager</i> property
     *
     * @return the newly created instance
     */
    public DataTransformerManager getDataTransformerManager() {
        String dataTransformerManager =
            singleton.getServerConfig().getEngineConfiguration().getDataTransformerManager();

        if (dataTransformerManager == null || dataTransformerManager.equals("")) {
            if (log.isLoggable(Level.FINEST)) {
                log.finest("DataTransformerManager not set");
            }
            return null;
        }
        try {
            return (DataTransformerManager)ConfigTools.getBeanInstance(
                classLoader,
                dataTransformerManager
            );
        } catch (Exception e) {
            if (log.isLoggable(Level.SEVERE)) {
                log.severe("Error creating the DataTransformerManager object: " + e.getMessage());
            }
            log.throwing(getClass().getName(), "getDataTransformerManager", e);
        }

        return null;
    }

    /**
     * Creates and returns a new Logging instance from the configured
     * <i>singleton.getServerConfig().getEngineConfiguration().logging</i> property
     *
     * @return the newly created instance
     */
    public LoggingConfiguration getLoggingConfiguration() {
        try {
            return (LoggingConfiguration)ConfigTools.getNewBeanInstance(
                classLoader,
                singleton.getServerConfig().getEngineConfiguration().getLoggingConfiguration()
            );
        } catch (Exception e) {
            if (log.isLoggable(Level.SEVERE)) {
                log.severe("Error creating the LoggingConfiguration object: " + e.getMessage());
            }
            log.throwing(getClass().getName(), "getLoggingConfiguration", e);
        }
        return null;
    }

    /**
     * Configures the logging system accordingly to the
     * com/funambol/server/logging/Logging.xml server bean.
     *
     * @throws IOException in case of errors setting up file handlers
     *
     */
    public void setLoggingConfiguration() throws IOException {

        //
        // First of all reinitialize logging
        //
        LogManager lm = LogManager.getLogManager();
        lm.reset();
        lm.readConfiguration();

        //
        // Now we can reconfigure Funambol logging
        //
        Logger l;
        LoggerConfiguration conf;
        LoggingConfiguration lc = getLoggingConfiguration();

        Handler h;
        Iterator i = lc.getLoggers().iterator();
        while(i.hasNext()) {

            conf = (LoggerConfiguration)i.next();

            //
            // If this logger inherits from its parent, just skip logger
            // configuration.
            //
            if (conf.isInherit()) {
                continue;
            }

            l = Logger.getLogger(conf.getName());

            //
            // first remove all old handlers
            //
            Handler[] handlers = l.getHandlers();
            for (int j=0; ((handlers != null) && (j<handlers.length)); ++j) {
                handlers[j].flush();
                handlers[j].close();
                l.removeHandler(handlers[j]);
            }
            l.setUseParentHandlers(false);

            //
            // Now process the new handlers.
            // Note that at the handler stage we set level to ALL (use the
            // logger level to select the logging granularity.
            //
            if (conf.isConsoleOutput()) {
                h = new ConsoleHandler();
                h.setFormatter(new com.funambol.framework.logging.SimpleFormatter());
                h.setLevel(Level.ALL);
                l.addHandler(h);
            }

            if (conf.isFileOutput()) {
                String pattern = conf.getPattern();

                if (new File(pattern).isAbsolute() == false) {
                    pattern = getSync4jHome() + File.separator + pattern;
                }

                if ((pattern == null) || (pattern.trim().length() == 0)) {
                    pattern = "logs/funambol_ds%t.log";
                }
                h = new FileHandler(
                    pattern        ,
                    conf.getLimit()*1048576,
                    conf.getCount(),
                    conf.isAppend()
                );
                h.setFormatter(new com.funambol.framework.logging.SimpleFormatter());
                h.setLevel(Level.ALL);
                l.addHandler(h);
            }

            if (LoggerConfiguration.LEVEL_OFF.equalsIgnoreCase(conf.getLevel())) {
                l.setLevel(Level.OFF);
            } else if (LoggerConfiguration.LEVEL_SEVERE.equalsIgnoreCase(conf.getLevel())) {
                l.setLevel(Level.SEVERE);
            } else if (LoggerConfiguration.LEVEL_INFO.equalsIgnoreCase(conf.getLevel())) {
                l.setLevel(Level.INFO);
            } else if (LoggerConfiguration.LEVEL_ALL.equalsIgnoreCase(conf.getLevel())) {
                l.setLevel(Level.ALL);
            }

        }
    }

    /**
     * Loads and instantiate a bean by its config name. In this case the bean
     * is not looked up in the internal configuration map, but is created
     * directly by. A new instance is create at each call.
     *
     * @param beanName the complete beanName
     *
     * @return the bean instance
     *
     * @throws BeanException in case of instantiation error
     */
    public Object getBeanInstanceByName(String beanName)
    throws BeanException {
        return getBeanInstanceByName(beanName, false);
    }

    /**
     * Loads and instantiate a bean by its config name. In this case the bean
     * is not looked up in the internal configuration map, but is created
     * directly by.
     *
     * If <code>cachedInstance</code> is true, this methods tries to return the
     * last instance created with the giveng server bean name. A new instance is
     * returned only when the server bean configuration has changed.
     *
     * @param beanName the complete beanName
     * @param cachedInstance should the instance be extracted from the cache?
     *
     * @return the bean instance
     *
     * @throws BeanException in case of instantiation error
     */
    public Object getBeanInstanceByName(String beanName, boolean cachedInstance)
    throws BeanException {
        if (cachedInstance) {
            return ConfigTools.getBeanInstance(classLoader, beanName);
        }

        return ConfigTools.getNewBeanInstance(classLoader, beanName);
    }

    /**
     * Saves an instantiate of a server bean to the given name under the config
     * path.
     *
     * @param beanName the complete beanName
     * @param obj the bean instance
     *
     *
     * @throws BeanException in case of instantiation error
     */
    public void setBeanInstance(String beanName, Object o)
    throws BeanException {
        ConfigTools.saveBeanInstance(beanName, new File(getConfigPath(), beanName), o);
    }

    /**
     * Returns the Funambol home as set as follows (the first value found is
     * taken):
     * <ul>
     *  <li>the system property funambol.ds.home</li>
     *  <li>the current directory</i>
     * </ul>
     *
     * @return the Funambol home as set as discribed above.
     */
    public static String getSync4jHome() {
        return System.getProperty(PROPERTY_SYSTEM_FUNAMBOL_DS_HOME, ".");
    }

    /**
     * Returns the config path calculated as getSync4jHome() + CONFIG_PATH
     *
     * @return the config path calculated as getSync4jHome() + CONFIG_PATH
     */
    public static String getConfigPath() {
        return getSync4jHome() + CONFIG_PATH;
    }

    /**
     * Redefined to print an understandable view of the singleton object.
     *
     * @return the string representing the singleton instance
     */
    public String toString() {
        StringBuffer sb = new StringBuffer();

        ServerConfiguration serverConfig = singleton.getServerConfig();

        sb.append(getClass().getName()).append('\n')
          .append("serverInfo: ").append(serverConfig.getServerInfo()).append('\n')
          .append("engineConfiguration: ").append(serverConfig.getEngineConfiguration()).append('\n')
          .append("classLoader: " + classLoader).append('\n');

        return sb.toString();
    }
}
