/**
 * Copyright (C) 2005-2006 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the Honest Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Honest Public License for more details.
 *
 * You should have received a copy of the Honest Public License
 * along with this program; if not, write to Funambol,
 * 643 Bair Island Road, Suite 305 - Redwood City, CA 94063, USA
 */
package com.funambol.server.notification.listener;

import java.io.*;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.servlet.ServletException;
import javax.servlet.ServletRequest;
import javax.servlet.http.*;

import org.jibx.runtime.*;

import com.funambol.framework.core.*;
import com.funambol.framework.logging.Sync4jLogger;
import com.funambol.framework.logging.Sync4jLoggerName;
import com.funambol.framework.protocol.ProtocolException;
import com.funambol.framework.security.Officer;
import com.funambol.framework.server.Sync4jDevice;
import com.funambol.framework.server.error.ServerException;
import com.funambol.framework.server.inventory.DeviceInventory;
import com.funambol.framework.server.inventory.DeviceInventoryException;
import com.funambol.framework.tools.IOTools;
import com.funambol.framework.tools.WBXMLTools;

import com.funambol.server.config.Configuration;
import javax.servlet.ServletConfig;

/**
 * Role of this servlet is to receive an address notification message,
 * authenticate the device and if the authentication is successful, register
 * the given address in the DeviceInventory
 *
 * @see com.funambol.framework.server.DeviceInventory
 *
 * @version $Id: Sync4jAddressListener.java,v 1.1 2006/08/08 07:51:02 nichele Exp $
 */
public class Sync4jAddressListener extends HttpServlet {

    // ------------------------------------------------------------ Private data
    private Logger log = Sync4jLogger.getLogger(Sync4jLoggerName.SERVER_NOTIFICATION);

    private static final String SOURCEURI_IP     = "ip"    ;
    private static final String SOURCEURI_MSISDN = "msisdn";

    private static final int SIZE_INPUT_BUFFER   = 1024;

    private static final String PARAM_DEFAULT_NOTIFICATION_BUILDER =
            "default-notification-builder";

    private static final String PARAM_DEFAULT_TCPIP_NOTIFICATION_SENDER =
            "default-tcpip-notification-sender";

    private static final String PARAM_DEFAULT_WAP_NOTIFICATION_SENDER =
            "default-wap-notification-sender";

    private String defaultNotificationBuilder      = null;
    private String defaultTCPIPNotificationSender  = null;
    private String defaultWAPNotificationSender    = null;

    // ---------------------------------------------------------- Public methods

    public void init() throws ServletException {

        ServletConfig config = getServletConfig();

        defaultNotificationBuilder =
            config.getInitParameter(PARAM_DEFAULT_NOTIFICATION_BUILDER);

        defaultTCPIPNotificationSender  =
            config.getInitParameter(PARAM_DEFAULT_TCPIP_NOTIFICATION_SENDER);

        defaultWAPNotificationSender  =
            config.getInitParameter(PARAM_DEFAULT_WAP_NOTIFICATION_SENDER);
    }


    public void doPost(HttpServletRequest request, HttpServletResponse response)
    throws ServletException, IOException {

        if (log.isLoggable(Level.FINEST)) {
            showHeaders(request);
        }

        if (log.isLoggable(Level.INFO)) {
            log.info("Handling incoming request for address listener.");
        }

        //
        //Read the message
        //
        byte[] binaryData = null;
        String xmlData    = null;

        InputStream in = request.getInputStream();
        binaryData     = IOTools.readContent(in, SIZE_INPUT_BUFFER);

        String ipRequest = ((ServletRequest)request).getRemoteAddr();

        try {
            if (in != null) {
                in.close();
            }
        } catch (Exception e) {
            log.severe(e.getMessage());
            log.throwing(getClass().getName(), "doPost", e.getCause());
            return;
        }
        in = null;

        final String contentType = request.getContentType().split(";")[0];
        try {
            if (Constants.MIMETYPE_SYNCMLDS_WBXML.equals(contentType)) {

                //
                // Convert WBXML into XML
                //
                if (log.isLoggable(Level.FINEST)) {
                    log.finest("Converting message from wbxml to xml");
                }
                xmlData = WBXMLTools.wbxmlToXml(binaryData);

            } else if (Constants.MIMETYPE_SYNCMLDS_XML.equals(contentType)) {
                //
                //None conversion is necessary
                //
                xmlData = new String(binaryData);
            } else {
                throw new ProtocolException( "Mime type "
                                           + contentType
                                           + " nor supported or unknown" );
            }
        } catch (Exception e) {
            log.severe(e.getMessage());
            log.throwing(getClass().getName(), "doPost", e.getCause());
            throw new ServletException(e);
        }

        //
        //Transform the string message into a SyncML object
        //
        if (log.isLoggable(Level.FINEST)) {
            log.finest("Message to translate into the SyncML object:\n" + xmlData);
        }

        SyncML syncObj = null;
        try {
            syncObj = convert(xmlData);
        } catch (Exception e) {
            log.severe("Unable to parse the request (" + e.getMessage() + ")");
            log.throwing(getClass().getName(), "doPost", e.getCause());
            response.setStatus(500);
            return;
        }

        if (log.isLoggable(Level.FINEST)) {
            log.finest("About processing message: " + Util.toXML(syncObj));
        }

        //
        //Authenticate device
        //
        boolean valid = false;
        if (syncObj.getSyncHdr().getCred() == null) {
            log.severe("Device not authenticated! Credentials are missing.");
            response.setStatus(407);
            return;
        }

        Cred   cred     = syncObj.getSyncHdr().getCred();
        String deviceId = syncObj.getSyncHdr().getSource().getLocURI();

        cred.getAuthentication().setDeviceId(deviceId);

        valid = authenticate(cred);
        if (!valid) {
            log.severe("Device not authenticated! Invalid credentials.");
            response.setStatus(401);
        } else {

            String deviceID = syncObj.getSyncHdr().getSource().getLocURI();

            DeviceInventory devInv = (DeviceInventory)Configuration
                                                     .getConfiguration()
                                                     .getDeviceInventory();

            Sync4jDevice device = new Sync4jDevice();
            device.setDeviceId(deviceID);
            try {
                //
                //Get device without capabilities
                //
                boolean found = devInv.getDevice(device, false);
                if (!found) {
                    response.setStatus(404);
                    return;
                }

            } catch (DeviceInventoryException e) {
                log.severe(e.getMessage());
                log.throwing(getClass().getName(), "doPost", e.getCause());

                response.setStatus(500);
                return;
            } catch (Exception e) {
                log.severe(e.getMessage());
                log.throwing(getClass().getName(), "doPost", e.getCause());

                response.setStatus(500);
                return;
            }
            //
            //This flag is used to tell if device's address or(and) msisdn
            //are specificated into Alert commands
            //
            boolean notified = false;

            boolean validIpAddressReceived = false;
            boolean msisdnReceived         = false;

            if (syncObj.getSyncBody().getCommands() != null) {
                int numCmds  = syncObj.getSyncBody().getCommands().size();
                int numItems = 0;
                ArrayList items = new ArrayList();
                String  sourceURI = null;
                String  data      = null;

                for (int i = 0; i < numCmds; i++) {
                    Alert alert = (Alert) syncObj.getSyncBody().getCommands().get(i);

                    //
                    //Check message data id - must be 745 to be processed
                    //
                    if (alert.getData() == 745) {
                        //
                        //Update device with informations contains into Alert
                        //
                        if (alert.getItems() != null) {
                            items    = alert.getItems();
                            numItems = items.size();
                            for (int j = 0; j < numItems; j++) {
                                Item item = (Item)items.get(j);
                                if (item.getSource()             != null &&
                                    item.getSource().getLocURI() != null &&
                                    item.getData()               != null   ) {

                                    if (log.isLoggable(Level.FINEST)) {
                                        log.finest("Item source URI : "  + item.getSource().getLocURI());
                                        log.finest("Item source DATA: "  + item.getData().getData());
                                    }

                                    sourceURI = item.getSource().getLocURI();
                                    if (sourceURI.equalsIgnoreCase(SOURCEURI_IP)) {
                                        data = item.getData().getData();
                                        //
                                        // Checks if the ip notified is the same
                                        // of the http request
                                        //
                                        if (ipRequest.equalsIgnoreCase(getIP(data))) {
                                            device.setAddress(data);
                                            //
                                            // We have received a valid alert with
                                            // the device ip, so if the device hasn't
                                            // a sender, we can set it to the default
                                            // value of the TCPIPSender (see init method)
                                            //
                                            validIpAddressReceived = true;
                                        } else {
                                            device.setAddress("");
                                        }
                                        notified = true;
                                    }
                                    if (sourceURI.equalsIgnoreCase(SOURCEURI_MSISDN)) {
                                        device.setMsisdn(item.getData().getData());
                                        msisdnReceived = true;
                                        notified       = true;
                                    }
                                }
                            }
                        }
                    }
                }

                if (notified) {
                    try {

                        if (validIpAddressReceived || msisdnReceived) {
                            //
                            // We have received a valid alert so if the device hasn't
                            // a notification builder, we set the default one.
                            //
                            if (device.getNotificationBuilder() == null ||
                                device.getNotificationBuilder().length() == 0) {
                                device.setNotificationBuilder(
                                        defaultNotificationBuilder);
                            }
                        }

                        if (validIpAddressReceived) {
                            //
                            // We have received a valid ip address. If the device
                            // hasn't a sender, we set the default tcp ip sender
                            //
                            if (device.getNotificationSender() == null ||
                                device.getNotificationSender().length() == 0) {
                                device.setNotificationSender(
                                        defaultTCPIPNotificationSender);
                            }
                        }

                        if (msisdnReceived) {
                            //
                            // We have received a msisdn. If the device
                            // hasn't a sender, we set the default wap sender
                            //
                            if (device.getNotificationSender() == null ||
                                device.getNotificationSender().length() == 0) {
                                device.setNotificationSender(
                                        defaultWAPNotificationSender);
                            }
                        }
                        devInv.setDevice(device);
                    } catch (DeviceInventoryException e) {
                        if (log.isLoggable(Level.SEVERE)) {
                            log.severe(e.getMessage());
                        }
                        log.throwing(getClass().getName(), "doPost", e.getCause());

                        response.setStatus(500);
                        return;
                    }

                    if (validIpAddressReceived || msisdnReceived) {
                        response.setStatus(200);
                    } else {
                        //
                        // We have received a valid Alert command (notified == true)
                        // but the ip isn't valid and there isn't the msisdn.
                        //
                        //
                        // Set status = 420 if there is an ip address includes
                        // into Alert commands that is different from the ip
                        // address of the client that sent the request
                        //
                        if (log.isLoggable(Level.FINEST)) {
                            log.finest("The IP address of the client is " +
                                       "different of the ip contained into Alert");
                        }
                        response.setStatus(420);
                    }
                } else {
                    if (log.isLoggable(Level.FINEST)) {
                        log.finest("Bad Request");
                    }
                    response.setStatus(400);
                }
            } else {
                log.severe("Message sent to AddressListener is null or " +
                           "does not contain any Alert command");
                response.setStatus(500);
            }
        }
    }

    // --------------------------------------------------------- Private methods

    /**
     * Authenticates a device
     *
     * @param credentials the credentials used for authentication
     *
     * @return true if device was successfuly authenticated, false otherwise
     * @throws IOException
     */
    private boolean authenticate(Cred credentials) throws IOException {
        Officer officer = Configuration.getConfiguration().getOfficer();
        return officer.authenticate(credentials);
    }

    /**
     * Converts a SyncML message into a SyncML object
     *
     * @param msg the message to be converted
     *
     * @return the SyncML object obtained from message
     * @throws ServerException
     */
    private SyncML convert(final String msg) throws ServerException {
        try {
            IBindingFactory f = BindingDirectory.getFactory(SyncML.class);
            IUnmarshallingContext c = f.createUnmarshallingContext();
            Object syncML = c.unmarshalDocument(new StringReader(msg));
            return (SyncML) syncML;
        } catch (JiBXException e) {
            if (log.isLoggable(Level.SEVERE)) {
                log.severe("Error unmarshalling message: " + e.getMessage());
            }
            log.throwing(getClass().getName(), "convert", e);
            throw new ServerException(e);
        }
    }

    /**
     * Log the headers of the request
     * @param request HttpServletRequest
     */
    private void showHeaders(HttpServletRequest request) {
        Enumeration enumHeaders = request.getHeaderNames();
        String headerName = null;
        log.finest("Http header: ");
        while (enumHeaders.hasMoreElements()) {
            headerName = (String)enumHeaders.nextElement();
            log.finest(headerName + ": " + request.getHeader(headerName));
        }
    }

    /**
     * Given an address like ip:port returns just the ip
     * @param address String
     * @return String
     */
    private String getIP(String address) {
        if (address == null) {
            return null;
        }
        int index = address.indexOf(":");
        if (index != -1) {
            return address.substring(0, index);
        }
        return address;
    }

}
