/**
 * Copyright (C) 2005-2006 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the Honest Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Honest Public License for more details.
 *
 * You should have received a copy of the Honest Public License
 * along with this program; if not, write to Funambol,
 * 643 Bair Island Road, Suite 305 - Redwood City, CA 94063, USA
 */
package com.funambol.server.notification.sender;

import java.io.*;

import java.net.InetSocketAddress;
import java.net.Socket;
import java.net.SocketAddress;

import java.util.logging.Level;
import java.util.logging.Logger;

import com.funambol.framework.logging.Sync4jLogger;
import com.funambol.framework.logging.Sync4jLoggerName;

import com.funambol.framework.notification.NotificationException;
import com.funambol.framework.notification.sender.Sender;

import com.funambol.framework.server.Sync4jDevice;

import com.funambol.framework.tools.DbgTools;


/**
 * This class is an implementation of Sender interface. It sends the
 * notification message over TCP/IP.
 *
 * @see Sender
 *
 * @version  $Id: TCPIPSender.java,v 1.1 2006/08/08 07:50:48 nichele Exp $
 */
public class TCPIPSender implements Sender {

    // --------------------------------------------------------------- Constants

    // ------------------------------------------------------------ Private data
    private Logger log = null;

    // -------------------------------------------------------------- Properties

    /** Default port used if no port is provided */
    private int  defaultPort = 745;

    public int getDefaultPort() {
        return this.defaultPort;
    }

    public void setDefaultPort(int port) {
        this.defaultPort = port;
    }

    /** Default connection timeout in seconds*/
    private int timeout = 60 * 1000;  // 1 minute

    public int getTimeout() {
        return this.timeout;
    }

    public void setTimeout(int timeout) {
        this.timeout = timeout;
    }

    /** Number of attempts to perform */
    private int attempts = 0;

    public int getAttempts() {
        return this.attempts;
    }

    public void setAttempts(int attempts) {
        this.attempts = attempts;
    }

    /** Time to wait between attempts in seconds */
    private int waitingTime = 0;

    public int getWaitingTime() {
        return this.waitingTime;
    }

    public void setWaitingTime(int waitingTime) {
        this.waitingTime = waitingTime;
    }


    // ------------------------------------------------------------- Constructor
    public TCPIPSender() {
        log = Sync4jLogger.getLogger(Sync4jLoggerName.SERVER_NOTIFICATION);
    }

    // ---------------------------------------------------------- Public Methods

    /**
     * Sends the given message to the device with the given phone number.
     *
     * @param device the recipient of the message
     * @param message the message to be sent
     *
     * @throws NotificationException
     */
    public void sendNotificationMessage(Sync4jDevice device, byte[] message)
    throws NotificationException {

        String deviceAddress= device.getAddress();

        if (deviceAddress == null || deviceAddress.equals("")) {
            if (log.isLoggable(Level.INFO)) {
                log.info("Address not configured for device '" + device.getDeviceId() + "'");
            }
            return ;
        }

        if (log.isLoggable(Level.INFO)) {
            log.info("sending message: " + DbgTools.bytesToHex(message));
            log.info("to: " + deviceAddress);
        }

        //try to get the connection port from address
        String address;
        int port;
        if (deviceAddress.indexOf(":") == -1) {
            //use default Funambol port
            address = deviceAddress;
            port = defaultPort;
        } else {
            //get port from address
            address = deviceAddress.substring(0, deviceAddress.indexOf(":"));
            port = Integer.parseInt(
                       deviceAddress.substring(deviceAddress.indexOf(":") + 1)
                   );
        }

        if (log.isLoggable(Level.INFO)) {
            log.info("Try to notify '" + address + "' on port '" +
                       port + "'");
        }
        if (log.isLoggable(Level.FINEST)) {
            log.finest("Attempts: "      + attempts +
                       ", timeout: "     + timeout +
                       ", waitingTime: " + waitingTime);
        }

        Socket           sock = null;
        DataOutputStream out  = null;
        BufferedReader   in   = null;

        if (attempts == 0) {
            attempts = 1;
        }
        for (int i=1; i<=attempts; i++) {
            try {
                if (log.isLoggable(Level.FINEST)) {
                    log.finest("Try to connect to '" + address + "' on port '" +
                               port + "'");
                    log.finest("Attempt n. " + i);
                }

                //create a socket and connect to device
                sock = new Socket();
                SocketAddress sockAddress = new InetSocketAddress(address, port);

                sock.connect(sockAddress, timeout * 1000);

                out = new DataOutputStream(sock.getOutputStream());

                //send message through socket
                out.write(message);

                //flsuh the message in stream
                out.flush();

                //read status response
                in = new BufferedReader(
                        new InputStreamReader(sock.getInputStream())
                     );
                String response = in.readLine();

                String code = response.substring(0, 3);

                if (log.isLoggable(Level.INFO)) {
                    log.info("Response status code for notification message sent over TCP/IP :" +
                            code);
                }

                if (!code.equals("200")) {
                        throw new NotificationException(
                                "Status code received for notification message sent over TCP/IP: " +
                                response);
                } else {
                    if (log.isLoggable(Level.INFO)) {
                        log.info("Notification message to '" + address +
                                 ":" + port + "' sent");

                    }
                    return;
                }

            } catch (Exception e) {
                log.throwing("TCPIPSender", "sendNotificationMessage", e);

                if (i == attempts) {
                    if (log.isLoggable(Level.INFO)) {
                        log.info("Send notification to '" + address +
                                 ":" + port + "' failed");
                    }

                    log.throwing("TCPIPSender", "sendNotificationMessage", e);
                    throw new NotificationException(e.getMessage());

                } else {
                    if (log.isLoggable(Level.FINEST)) {
                        log.finest("Attempt n. " + i + " failed for '" + address +
                                   ":" + port + "'");
                        log.finest("Wait " + waitingTime + " seconds");

                    }
                    try {
                        Thread.currentThread().sleep(waitingTime * 1000);
                    } catch (Exception ex) {
                        if (log.isLoggable(Level.FINEST)) {
                            log.finest("Error waiting between the attempts...");

                        }

                        log.throwing("TCPIPSender", "sendNotificationMessage", e);
                        throw new NotificationException(ex.getMessage());
                    }
                }
            } finally {

                //close connection to device
                try {
                    if (out != null) {
                        out.close();
                    }
                    if (in != null) {
                        in.close();
                    }
                    if (sock != null) {
                        if (!sock.isClosed()) {
                            sock.close();
                        }
                    }
                } catch (IOException e) {
                    if (log.isLoggable(Level.FINEST)) {
                        log.finest("Error closing connection");
                    }
                    log.throwing("TCPIPSender", "sendNotificationMessage", e);
                }
            }
        }
    }

}
