/**
 * Copyright (C) 2003-2006 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the Honest Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Honest Public License for more details.
 *
 * You should have received a copy of the Honest Public License
 * along with this program; if not, write to Funambol,
 * 643 Bair Island Road, Suite 305 - Redwood City, CA 94063, USA
 */
package com.funambol.server.security;

import java.security.Principal;
import java.util.logging.Level;
import java.util.logging.Logger;

import com.funambol.framework.core.Authentication;
import com.funambol.framework.core.Cred;
import com.funambol.framework.filter.Clause;
import com.funambol.framework.filter.LogicalClause;
import com.funambol.framework.filter.WhereClause;
import com.funambol.framework.logging.Sync4jLogger;
import com.funambol.framework.security.Officer;
import com.funambol.framework.security.Sync4jPrincipal;
import com.funambol.framework.server.Sync4jUser;
import com.funambol.framework.server.store.PersistentStore;
import com.funambol.framework.server.store.PersistentStoreException;
import com.funambol.framework.tools.Base64;
import com.funambol.framework.tools.MD5;
import com.funambol.framework.tools.beans.LazyInitBean;

import com.funambol.server.admin.UserManager;
import com.funambol.server.config.Configuration;


/**
 * This is an implementation of the <i>Officier</i> interface. It always
 * authenticates and authorizes users and resource accesses.
 *
 * @version $Id: DBOfficer.java,v 1.8 2006/08/07 21:13:17 nichele Exp $
 */
public class DBOfficer implements Officer, java.io.Serializable, LazyInitBean {

    // --------------------------------------------------------------- Constants
    protected static final String ROLE_USER = "sync_user";

    // ---------------------------------------------------------- Protected data

    protected transient Logger log    = Sync4jLogger.getLogger();
    protected PersistentStore ps      = null;
    protected UserManager userManager = null;

    // ------------------------------------------------------------- Public data
    // -------------------------------------------------------------- Properties

    /**
     * Has the last login failed for expired temporary login?
     */
    private boolean loginExpired = false;

    /**
     * Which type of authetication does impose the server?
     */
    private String clientAuth = Cred.AUTH_TYPE_BASIC;

    public String getClientAuth() {
        return this.clientAuth;
    }
    public void setClientAuth(String clientAuth) {
        this.clientAuth = clientAuth;
    }

    /**
     * Which type of authetication use the server to send his credential?
     */
    private String serverAuth = Cred.AUTH_NONE;
    public String getServerAuth() {
        return this.serverAuth;
    }

    public void setServerAuth(String serverAuth) {
        this.serverAuth = serverAuth;
    }

    // ------------------------------------------------------------ Constructors
    public DBOfficer() {
    }

    // ---------------------------------------------------------- Public methods

    public void init() {
        Configuration config = Configuration.getConfiguration();
        ps = config.getStore();

        userManager = (UserManager)config.getUserManager();
    }

    /**
     * Authenticates a credential.
     *
     * @param credential the credential to be authenticated
     *
     * @return true if the credential is autenticated, false otherwise
     */
    public boolean authenticate(Cred credential) {

        String type = credential.getType();

        if ((Cred.AUTH_TYPE_BASIC).equals(type)) {
            return authenticateBasicCredential(credential);
        } else if ( (Cred.AUTH_TYPE_MD5).equals(type)  ) {
            String syncMLVerProto =
                             credential.getAuthentication().getSyncMLVerProto();

            if (log.isLoggable(Level.FINEST)) {
                log.finest("Check MD5 credential with protocol " + syncMLVerProto);
             }

            //
            // The modality in order to calculate the digest is different to second
            // of the used version of SyncML
            //
            boolean isSyncML10 = false;
            if (syncMLVerProto.indexOf("1.0") != -1) {
                isSyncML10 = true;
            }
            return authenticateMD5Credential(credential, isSyncML10);
        } else if ( (Cred.AUTH_TYPE_HMAC).equals(type)  ) {
            return false;
        }
        return false;
    }

    /**
     * Authorizes a resource.
     *
     * @param principal the requesting entity
     * @param resource the name (or the identifier) of the resource to be authorized
     *
     * @return true if the credential is authorized to access the resource, false
     *         otherwise
     */
    public boolean authorize(Principal principal, String resource) {
        return true;
    }

    /** Un-authenticates a credential.
     *
     * @param credential the credential to be unauthenticated
     */
    public void unAuthenticate(Cred credential) {
        //
        // Do nothing. In the current implementation, the authentication is
        // discarde as soon as the LoginContext is garbage collected.
        //
    }

    /**
     * @see com.funambol.framework.security.Officer
     */
    public boolean isAccountExpired() {
        return loginExpired;
    }

    // ------------------------------------------------------- Protected Methods
    /**
     * Returns the user (with the own roles) with the given username and password
     * (or null if it is not found).
     * The password could be null and in this case the user with the given username
     * is returned (or null if it is not found)
     * @param userName the username
     * @param password the password
     * @return the user with the own roles
     */
    protected Sync4jUser getUser(String userName, String password) {

        Sync4jUser[] users = null;

        try {
            Clause      verifyUserClause = null;
            WhereClause usernameClause   = null;
            WhereClause passwordClause   = null;

            String value[] = new String[1];
            value[0] = userName;
            usernameClause = new WhereClause("username", value, WhereClause.OPT_EQ, true);

            if (password != null) {
                value = new String[1];
                value[0] = password;
                passwordClause = new WhereClause("password", value, WhereClause.OPT_EQ, true);
                verifyUserClause = new LogicalClause(LogicalClause.OPT_AND,
                                                     new WhereClause[] {usernameClause, passwordClause});
            } else {
                verifyUserClause = usernameClause;
            }

            users = userManager.getUsers(verifyUserClause);
            if (users.length == 1) {
                userManager.getUserRoles(users[0]);
                return users[0];
            }

        } catch (PersistentStoreException e) {
            log.severe("Error reading user: " + e);
            log.throwing(getClass().getName(), "verifyUser", e);
            return null;
        }

        return null;
    }

    /**
     * Checks if the given user is a sync_user
     * @param user the user to check
     * @return true if the given user is a sync_user, false otherwise
     */
    protected boolean isASyncUser(Sync4jUser user) {
        //
        // Check the roles
        //
        String[] roles = user.getRoles();
        if (roles == null || roles.length == 0) {
            return false;
        }
        for (int i = 0; i < roles.length; i++) {
            if (ROLE_USER.equals(roles[i])) {
                return true;
            }
        }
        return false;
    }

    // --------------------------------------------------------- Private Methods

    /**
     * Checks the given B64 credential.
     *
     * @param credential the cred to check
     * @return true is the cred are authenticate, false otherwise.
     */
    private boolean authenticateBasicCredential(Cred credential) {
        Sync4jUser[] users = null;
        String username = null, password = null;
        Sync4jPrincipal principal = null;

        Authentication auth = credential.getAuthentication();
        String deviceId     = auth.getDeviceId();
        String userpwd      = new String(Base64.decode(auth.getData()));

        int p = userpwd.indexOf(':');

        if (p == -1) {
            username = userpwd;
            password = "";
        } else {
            username = (p>0) ? userpwd.substring(0, p) : "";
            password = (p == (userpwd.length()-1)) ? "" : userpwd.substring(p+1);
        }

        if (log.isLoggable(Level.FINEST)) {
            log.finest("Username: " + username);
        }

        //
        // Verify the user
        //
        Sync4jUser user = getUser(username, password);

        if (user == null) {

            if (log.isLoggable(Level.FINEST)) {
                log.finest("User '" + username + "' not found.");
            }
            return false;

        } else {

            if (log.isLoggable(Level.FINEST)) {
                log.finest("User '" + username + "' found");
            }
            //
            // Check the roles
            //
            boolean isASyncUser = isASyncUser(user);

            if (isASyncUser) {
                //
                // User authenticate
                //
                if (log.isLoggable(Level.FINEST)) {
                    log.finest("User authenticate");
                }
            } else {
                //
                // User not authenticate
                //
                if (log.isLoggable(Level.FINEST)) {
                    log.finest("The user is not a '" + ROLE_USER + "'");
                }
                return false;
            }
        }

        //
        // Verify that the principal for the specify deviceId and username exists
        //
        try {
            principal = Sync4jPrincipal.createPrincipal(username, deviceId);
            ps.read(principal);
            credential.getAuthentication().setPrincipalId(principal.getId());

            return true;

        } catch(PersistentStoreException e) {
            log.severe("Error reading principal: " + e);
            log.throwing(getClass().getName(), "authenticateBasicCredential", e);
        }

        return false;
    }

    /**
     * Checks if the digest computed with the given clientNonce, userName, password
     * and the syncml 10 algorithm is equal to the given clientDigest.
     * @param clientDigest the client digest
     * @param clientNonce the nonce to use in the digest computation
     * @param userName the username
     * @param password the password
     * @return true if the clientDigest is right, false otherwise
     */
    private boolean checkMD5Credential10(String clientDigest,
                                         byte[] clientNonce,
                                         String userName,
                                         String password) {
        //
        // Calculate digest
        //
        String usernamepwd = userName + ':' + password;

        byte[] usernamepwdBytes = usernamepwd.getBytes();

        if (log.isLoggable(Level.FINEST)) {
            log.finest("username: " + userName);
            log.finest("password: " + password);
            log.finest("clientNonce  : " + new String(Base64.encode(clientNonce)));
        }

        //
        // computation of the MD5 digest
        //
        // Creates a unique buffer containing the bytes to digest
        //
        byte[] buf = new byte[usernamepwdBytes.length + 1 + clientNonce.length];

        System.arraycopy(usernamepwdBytes, 0, buf, 0, usernamepwdBytes.length);
        buf[usernamepwdBytes.length] = (byte) ':';
        System.arraycopy(clientNonce, 0, buf, usernamepwdBytes.length + 1,
                         clientNonce.length);

        byte[] digest = MD5.digest(buf);

        //
        // encoding digest in Base64 for comparation with digest sent by client
        //
        String serverDigestNonceB64 = new String(Base64.encode(digest));

        if (log.isLoggable(Level.FINEST)) {
            log.finest("serverDigestNonceB64: " + serverDigestNonceB64);
            log.finest("clientDigest: " + clientDigest);
        }

        return clientDigest.equals(serverDigestNonceB64);
    }

    /**
     * Checks if the digest computed with the given clientNonce, userName, password
     * and the syncml 11 algorithm is equal to the given clientDigest.
     * @param clientDigest the client digest
     * @param clientNonce the nonce to use in the digest computation
     * @param userName the username
     * @param password the password
     * @return true if the clientDigest is right, false otherwise
     */
    private boolean checkMD5Credential11(String clientDigest,
                                         byte[] clientNonce,
                                         String userName,
                                         String password) {
        //
        // Calculate digest
        //
        byte[] userDigest = MD5.digest((new String(userName + ':' + password)).
                                       getBytes());
        byte[] userDigestB64 = Base64.encode(userDigest);

        if (log.isLoggable(Level.FINEST)) {
            log.finest("username: " + userName);
            log.finest("userDigestB64: " + new String(userDigestB64));
            log.finest("clientNonce  : " + new String(Base64.encode(clientNonce)));
        }

        //
        // computation of the MD5 digest
        //
        // Creates a unique buffer containing the bytes to digest
        //
        byte[] buf = new byte[userDigestB64.length + 1 + clientNonce.length];

        System.arraycopy(userDigestB64, 0, buf, 0, userDigestB64.length);
        buf[userDigestB64.length] = (byte) ':';
        System.arraycopy(clientNonce, 0, buf, userDigestB64.length + 1,
                         clientNonce.length);

        byte[] digest = MD5.digest(buf);

        //
        // encoding digest in Base64 for comparation with digest sent by client
        //
        String serverDigestNonceB64 = new String(Base64.encode(digest));

        if (log.isLoggable(Level.FINEST)) {
            log.finest("serverDigestNonceB64: " + serverDigestNonceB64);
            log.finest("clientDigest: " + clientDigest);
        }

        return clientDigest.equals(serverDigestNonceB64);
    }

    /**
     * Checks the given MD5 credential using the appropriate algorithm in according
     * to the given parameter isSyncML10
     * @param credential the cred to check
     * @param isSyncML10 have we tou use the algo for SyncML 10 ?
     * @return true is the cred are authenticate, false otherwise.
     */
    private boolean authenticateMD5Credential(Cred credential, boolean isSyncML10) {

        Sync4jPrincipal principals[] = null;
        boolean isAuthenticate = false;

        String password = null;
        String userName = null;

        Authentication auth = credential.getAuthentication();
        String deviceId = auth.getDeviceId();

        //
        // Gets all principals for the specify deviceId
        //
        WhereClause whereClause = new WhereClause("device",
                                                  new String[] {deviceId},
                                                  WhereClause.OPT_EQ,
                                                  true);
        try {
            principals = (Sync4jPrincipal[]) ps.read(
                             Sync4jPrincipal.createPrincipal(null, deviceId),
                             whereClause
                         );
        } catch (PersistentStoreException e) {
            log.severe("Error reading principals: " + e);
            log.throwing(getClass().getName(), "authenticateMD5Credential", e);
            return false;
        }
        //
        // Search the username
        //
        int size = principals.length;
        Sync4jUser user = null;
        boolean isASyncUser = false;
        for (int i = 0; i < size; i++) {

            user = getUser(principals[i].getUsername(), null);
            isASyncUser = isASyncUser(user);

            if (!isASyncUser) {
                //
                // Skip this user, he is not a sync_user
                //
                continue;
            }

            userName = user.getUsername();
            password = user.getPassword();

            //
            // digest sent by client in format b64
            //
            String msgDigestNonceB64 = auth.getData();
            byte[] clientNonce = auth.getNextNonce().getValue();

            boolean digestOk = false;

            if (isSyncML10) {
                digestOk = checkMD5Credential10(msgDigestNonceB64,
                                                clientNonce,
                                                userName,
                                                password);
            } else {
                digestOk = checkMD5Credential11(msgDigestNonceB64,
                                                clientNonce,
                                                userName,
                                                password);
            }
            if (digestOk) {
                credential.getAuthentication().setPrincipalId(principals[i].getId());
                credential.getAuthentication().setUsername(userName);
                isAuthenticate = true;
                break;
            }
        }

        if (!isAuthenticate) {
            return false;
        }

        credential.setAuthentication(auth);
        return true;
    }

}
