/**
 * Copyright (C) 2006 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the Honest Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Honest Public License for more details.
 *
 * You should have received a copy of the Honest Public License
 * along with this program; if not, write to Funambol,
 * 643 Bair Island Road, Suite 305 - Redwood City, CA 94063, USA
 */
package com.funambol.server.security;

import java.security.Principal;
import java.util.logging.Level;

import com.funambol.framework.core.Authentication;
import com.funambol.framework.core.Cred;
import com.funambol.framework.security.Sync4jPrincipal;
import com.funambol.framework.server.Sync4jUser;
import com.funambol.framework.server.store.NotFoundException;
import com.funambol.framework.server.store.PersistentStoreException;
import com.funambol.framework.tools.Base64;
import com.funambol.framework.tools.beans.LazyInitBean;

import com.funambol.server.admin.UserManager;
import com.funambol.server.config.Configuration;


/**
 * This is an implementation of the <i>Officier</i> interface. It provides
 * the user provisioning so if an user is not in the database he will be added.
 * It requires basic authentication
 *
 * @version $Id: UserProvisioningOfficer.java,v 1.1 2006/08/08 07:53:01 nichele Exp $
 */
public class UserProvisioningOfficer extends DBOfficer implements LazyInitBean {

    // ---------------------------------------------------------- Protected data
    // ------------------------------------------------------------- Public data
    // -------------------------------------------------------------- Properties

    /**
     * Has the last login failed for expired temporary login?
     */
    private boolean loginExpired = false;

    /**
     * Which type of authetication does impose the server?
     */
    private String clientAuth = Cred.AUTH_TYPE_BASIC;

    public String getClientAuth() {
        return this.clientAuth;
    }

    /**
     * Which type of authetication use the server to send his credential?
     */
    private String serverAuth = Cred.AUTH_NONE;

    public String getServerAuth() {
        return this.serverAuth;
    }

    public void setServerAuth(String serverAuth) {
        this.serverAuth = serverAuth;
    }

    // ------------------------------------------------------------ Constructors
    public UserProvisioningOfficer() {
        super();
    }

    // ---------------------------------------------------------- Public methods

    public void init() {
        super.init();
    }

    /**
     * Authenticates a credential.
     *
     * @param credential the credential to be authenticated
     *
     * @return true if the credential is autenticated, false otherwise
     */
    public boolean authenticate(Cred credential) {

        Configuration config = Configuration.getConfiguration();
        ps = config.getStore();

        userManager     = (UserManager)config.getUserManager();

        String type = credential.getType();

        if ((Cred.AUTH_TYPE_BASIC).equals(type)) {
            return authenticateBasicCredential(credential);
        }

        return false;
    }

    /**
     * Authorizes a resource.
     *
     * @param principal the requesting entity
     * @param resource the name (or the identifier) of the resource to be authorized
     *
     * @return true if the credential is authorized to access the resource, false
     *         otherwise
     */
    public boolean authorize(Principal principal, String resource) {
        return true;
    }

    /** Un-authenticates a credential.
     *
     * @param credential the credential to be unauthenticated
     */
    public void unAuthenticate(Cred credential) {
        //
        // Do nothing. In the current implementation, the authentication is
        // discarde as soon as the LoginContext is garbage collected.
        //
    }

    /**
     * @see com.funambol.framework.security.Officer
     */
    public boolean isAccountExpired() {
        return loginExpired;
    }

    // ------------------------------------------------------- Protected Methods
    /**
     * Checks the given credential. If the user or the principal isn't found,
     * they are created.
     * @param credential the credential to check
     * @return boolean true if no errors occur.
     */
    protected boolean authenticateBasicCredential(Cred credential) {
        String username = null, password = null;

        Sync4jPrincipal principal = null;

        Authentication auth = credential.getAuthentication();
        String deviceId = auth.getDeviceId();

        String userpwd = new String(Base64.decode(auth.getData()));

        int p = userpwd.indexOf(':');

        if (p == -1) {
            username = userpwd;
            password = "";
        } else {
            username = (p > 0) ? userpwd.substring(0, p) : "";
            password = (p == (userpwd.length() - 1)) ? "" :
                       userpwd.substring(p + 1);
        }

        if (log.isLoggable(Level.FINEST)) {
            log.finest("User to check: " + username);
        }

        //
        // Gets the user without checking the password
        //
        Sync4jUser user = getUser(username, null);

        if (user == null) {

            if (log.isLoggable(Level.FINEST)) {
                log.finest("User '" + username + "' not found. A new user will be created");
            }

            try {
                insertUser(username, password);
                if (log.isLoggable(Level.FINEST)) {
                    log.finest("User '" + username + "' created");
                }
            } catch (PersistentStoreException e) {
                log.severe("Error inserting a new user: " + e);
                log.throwing(getClass().getName(),
                             "authenticateBasicCredential", e);
                return false;
            }

        } else {
            if (log.isLoggable(Level.FINEST)) {
                log.finest("User '" + username + "' found");
            }
            //
            // Check the password
            //
            String storedPassword = user.getPassword();
            if (!password.equals(storedPassword)) {
                //
                // The user isn't authenticate
                //
                if (log.isLoggable(Level.FINEST)) {
                    log.finest("The sent password is different from the stored one. User not authenticate");
                }
                return false;
            } else {
                //
                // Check the roles
                //
                boolean isASyncUser = isASyncUser(user);

                if (isASyncUser) {
                    //
                    // User authenticate
                    //
                    if (log.isLoggable(Level.FINEST)) {
                        log.finest("User authenticate");
                    }
                } else {
                    //
                    // User not authenticate
                    //
                    if (log.isLoggable(Level.FINEST)) {
                        log.finest("The user is not a '" + ROLE_USER + "'");
                    }
                    return false;
                }
            }
        }

        //
        // Verify that the principal for the specify deviceId and username exists
        // Otherwise a new principal will be created
        //
        try {
            principal = handlePrincipal(username, deviceId);
        } catch (PersistentStoreException e) {
            if (log.isLoggable(Level.SEVERE)) {
                log.severe("Error handling the principal: " + e);
            }
            log.throwing(getClass().getName(), "authenticateBasicCredential", e);
            return false;
        }

        credential.getAuthentication().setPrincipalId(principal.getId());

        return true;
    }

    /**
     * Insert a new user with the given username and password
     * @param userName the username
     * @param password the password
     * @throws PersistentStoreException if an error occurs
     */
    protected void insertUser(String userName, String password)
    throws PersistentStoreException {

        Sync4jUser user = new Sync4jUser();
        user.setUsername(userName);
        user.setPassword(password);
        user.setRoles(new String[] {ROLE_USER});
        userManager.insertUser(user);

    }

    /**
     * Returns the principal with the given username and deviceId.
     * <code>null</code> if not found
     * @param userName the username
     * @param deviceId the device id
     * @return the principal found or null.
     */
    protected Sync4jPrincipal getPrincipal(String userName, String deviceId) throws
            PersistentStoreException {

        Sync4jPrincipal principal = null;

        //
        // Verify that exist the principal for the specify deviceId and username
        //

        principal = Sync4jPrincipal.createPrincipal(userName, deviceId);

        try {
            ps.read(principal);
        } catch (NotFoundException ex) {
            return null;
        }

        return principal;
    }

    /**
     * Inserts a new principal with the given userName and deviceId
     * @param userName the username
     * @param deviceId the device id
     * @return the principal created
     * @throws PersistentStoreException if an error occurs creating the principal
     */
    protected Sync4jPrincipal insertPrincipal(String userName, String deviceId) throws
            PersistentStoreException {

        //
        // We must create a new principal
        //
        Sync4jPrincipal principal = Sync4jPrincipal.createPrincipal(userName, deviceId);

        ps.store(principal);

        return principal;
    }


    /**
     * Searchs if there is a principal with the given username and device id.
     * if no principal is found, a new one is created.
     * @param userName the user name
     * @param deviceId the device id
     * @return the found principal or the new one
     */
    protected Sync4jPrincipal handlePrincipal(String username, String deviceId)
    throws PersistentStoreException {

        Sync4jPrincipal principal = null;

        //
        // Verify if the principal for the specify deviceId and username exists
        //

        principal = getPrincipal(username, deviceId);

        if (log.isLoggable(Level.FINEST)) {
            log.finest("Principal '" + username +
                       "/" +
                       deviceId + "' " +
                       ((principal != null) ? "found" : "not found. A new principal will be created"));
        }

        if (principal == null) {
            principal = insertPrincipal(username, deviceId);
            if (log.isLoggable(Level.FINEST)) {
                log.finest("Principal '" + username +
                           "/" +
                           deviceId + "' created");
            }
        }

        return principal;
    }
}
