/**
 * Copyright (C) 2003-2006 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the Honest Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Honest Public License for more details.
 *
 * You should have received a copy of the Honest Public License
 * along with this program; if not, write to Funambol,
 * 643 Bair Island Road, Suite 305 - Redwood City, CA 94063, USA
 */
package com.funambol.server.store;

import java.util.HashMap;

import java.sql.*;

import com.funambol.framework.filter.*;

import com.funambol.framework.server.ClientMapping;

import com.funambol.framework.server.store.PersistentStore;
import com.funambol.framework.server.store.BasePersistentStore;
import com.funambol.framework.server.store.PersistentStoreException;

import com.funambol.framework.tools.DBTools;
import java.util.ArrayList;
import java.util.List;
import com.funambol.framework.server.ClientMappingEntry;

/**
 * This is the store for information regarding the synchronization process and
 * status. It persists the following classes:
 *
 * <ul>
 * <li>com.funambol.framework.server.ClientMapping</li>
 * </ul>
 *
 * This <i>PersistentStore</i> is configured with the following map key:
 * <ul>
 *  <li>jndiDataSourceName</li>
 * </ul>
 *
 * @version $Id: ClientMappingPersistentStore.java,v 1.3 2006/08/07 21:13:17 nichele Exp $
 */
public class ClientMappingPersistentStore
extends    BasePersistentStore
implements PersistentStore, java.io.Serializable {

    // --------------------------------------------------------------- Constants
    public static final int SQL_INSERT_CLIENT_MAPPING = 0;
    public static final int SQL_UPDATE_CLIENT_MAPPING = 1;
    public static final int SQL_SELECT_CLIENT_MAPPING = 2;
    public static final int SQL_DELETE_CLIENT_MAPPING = 3;

    // -------------------------------------------------------------- Properties

    private String[] sql = null;

    public void setSql(String[] sql) {
        this.sql = sql;
    }

    public String[] getSql() {
        return this.sql;
    }

    // ------------------------------------------------------------ Constructors

    // ---------------------------------------------------------- Public methods

    public boolean store(Object o)
    throws PersistentStoreException {
        if (o instanceof ClientMapping) {
            ClientMapping clientMapping = (ClientMapping) o;

            Connection conn = null;
            PreparedStatement stmt = null, stmtIns = null;

            long principal = -1;
            String dbURI     = clientMapping.getDbURI();

            principal = clientMapping.getPrincipal().getId();

            assert ((principal >= 0) && (dbURI != null));

            try {
                conn = dataSource.getConnection();

                if (clientMapping.isDeleted()) {
                    stmt = conn.prepareStatement(sql[SQL_DELETE_CLIENT_MAPPING]);
                    stmt.setLong(1, principal);
                    stmt.setString(2, dbURI    );
                    String[] luids = clientMapping.getDeletedLuids();
                    for (int i=0; i < luids.length; ++i) {
                        stmt.setString(3, luids[i]);
                        stmt.executeUpdate();
                    }
                    stmt.close(); stmt = null;
                }

                if (clientMapping.isModified()) {
                    stmt = conn.prepareStatement(sql[SQL_UPDATE_CLIENT_MAPPING]);
                    stmt.setLong  (3, principal);
                    stmt.setString(4, dbURI    );
                    String[] luids = clientMapping.getModifiedLuids();

                    String guid       = null;
                    String lastAnchor = null;
                    for (int i=0; i < luids.length; ++i) {
                        int n = -1;

                        guid       = clientMapping.getMappedValueForLuid(luids[i]);
                        lastAnchor = clientMapping.getLastAnchor(guid);

                        stmt.setString(1, luids[i]  );
                        stmt.setString(2, lastAnchor);
                        stmt.setString(5, guid      );
                        n = stmt.executeUpdate();

                        if (n == 0) {
                            //
                            // insert new mapping
                            //
                            stmtIns = conn.prepareStatement(sql[SQL_INSERT_CLIENT_MAPPING]);
                            stmtIns.setLong  (1, principal);
                            stmtIns.setString(2, dbURI);
                            stmtIns.setString(3, guid);
                            stmtIns.setString(4, luids[i]);
                            if (lastAnchor == null) {
                                stmtIns.setNull  (5, Types.VARCHAR);
                            } else {
                                stmtIns.setString(5, lastAnchor);
                            }
                            stmtIns.executeUpdate();
                            stmtIns.close(); stmtIns = null;
                        }
                    }
                    stmt.close(); stmt = null;
                }
            } catch (SQLException e) {
                throw new PersistentStoreException("Error storing client mapping", e);
            } finally {
                DBTools.close(conn, stmt, null);
            }
            return true;
        }
        return false;
    }

    public boolean read(Object o)
    throws PersistentStoreException {
        if (o instanceof ClientMapping) {
            ClientMapping clientMapping = (ClientMapping) o;

            Connection conn = null;
            PreparedStatement stmt = null;
            ResultSet rs = null;

            try {
                conn = dataSource.getConnection();

                stmt = conn.prepareStatement(sql[SQL_SELECT_CLIENT_MAPPING]);
                stmt.setLong  (1, clientMapping.getPrincipal().getId());
                stmt.setString(2, clientMapping.getDbURI());
                rs = stmt.executeQuery();

                List               mapping      = new ArrayList();
                ClientMappingEntry mappingEntry = null;

                while (rs.next()) {
                    mappingEntry = new ClientMappingEntry(rs.getString("guid"),
                                                          rs.getString("luid"),
                                                          rs.getString("last_anchor"));
                    mapping.add(mappingEntry);
                }
                clientMapping.initializeFromMapping(mapping);
            } catch (SQLException e) {
                throw new PersistentStoreException("Error reading mapping", e);
            } finally {
                DBTools.close(conn, stmt, rs);
            }
            return true;
        }
        return false;
    }

    /** Read all objects stored the persistent media.
     *
     * @param objClass the object class handled by the persistent store
     *
     * @return an array containing the objects read. If no objects are found an
     *         empty array is returned. If the persistent store has not
     *         processed the quest, null is returned.
     *
     * @throws PersistentStoreException
     *
     */
    public Object[] read(Class objClass) throws PersistentStoreException {
        //
        // TO DO (not used yet)
        //
        return null;
    }

    public boolean delete(Object o) throws PersistentStoreException
    {
        return false;
    }

    public Object[] read(Object o, Clause clause) throws PersistentStoreException
    {
        return null;
    }

    public int count(Object o, Clause clause) throws PersistentStoreException
    {
        return -1;
    }

    public boolean store(Object o, String operation) throws PersistentStoreException
    {
        return false;
    }

}
