/**
 * Copyright (C) 2003-2006 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the Honest Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Honest Public License for more details.
 *
 * You should have received a copy of the Honest Public License
 * along with this program; if not, write to Funambol,
 * 643 Bair Island Road, Suite 305 - Redwood City, CA 94063, USA
 */
package com.funambol.server.syncbean;

import javax.ejb.*;
import javax.naming.*;

import javax.rmi.PortableRemoteObject;

import java.io.*;
import java.util.Map;
import java.util.logging.Logger;
import java.util.logging.Level;

import com.funambol.framework.core.*;
import com.funambol.framework.server.SyncResponse;
import com.funambol.framework.logging.Sync4jLogger;
import com.funambol.framework.config.ConfigClassLoader;
import com.funambol.framework.config.ConfigurationException;
import com.funambol.framework.protocol.ProtocolException;

import com.funambol.framework.core.Sync4jException;
import com.funambol.framework.core.RepresentationException;
import com.funambol.framework.server.session.SessionExpiredException;
import com.funambol.framework.server.error.*;

import com.funambol.server.config.Configuration;
import com.funambol.server.engine.SyncAdapter;
import com.funambol.server.session.SyncSessionHandler;

import com.funambol.server.admin.*;
import com.funambol.server.admin.ejb.*;

/**
 *  This is the session enterprise java bean that handles a synchronization
 *  request. It is designed to be a stateful session bean so that the state
 *  and the session management are on charge of the application server.
 *
 *  LOG NAME: funambol.server
 *
 *  @version $Id: SyncBean.java,v 1.3 2006/08/07 21:13:17 nichele Exp $
 *
 */
public class SyncBean
implements javax.ejb.SessionBean {

    // ------------------------------------------------------- Private constants

    private static final String ENV_ADMIN_NAME
        = "java:comp/env/ejb/LocalAdminBean";

    // ------------------------------------------------------------ Private data

    private transient Logger log = null;

    private SessionContext sessionContext       = null;

    private SyncSessionHandler sessionHandler = null;

    private Configuration config                = null;

    private String sessionId                    = null;

    private SyncAdapter syncAdapter             = null;

    // ------------------------------------------------------------- EJB methods

    public void ejbCreate(String sessionId) throws CreateException {

        log = Sync4jLogger.getLogger("server");

        loadAdminBean();

        try {
            syncAdapter = new SyncAdapter(config);
            syncAdapter.beginSync(sessionId);
        } catch (Exception e) {
            log.throwing("SyncBean", "ejbCreate", e);
            throw new CreateException( "Error "
                                     + e.getClass().getName()
                                     + " creating the SyncBean: "
                                     + e.getMessage()
                                     );
        }
        this.sessionId = sessionId;
    }

    /**
     * Creates an instance of AdminBean
     */
    private void loadAdminBean() throws CreateException {

        try {

            InitialContext ctx = new InitialContext();

            Object obj = ctx.lookup(ENV_ADMIN_NAME);
            AdminHomeLocal home =
                (AdminHomeLocal)PortableRemoteObject.narrow(obj, AdminHomeLocal.class);
            AdminLocal admin = home.create();

            config = admin.getConfig();

        } catch (NamingException e) {
            e.printStackTrace();
            log.throwing("SyncBean", "loadAdminBean", e);
            throw new CreateException( "Error "
                                     + e.getClass().getName()
                                     + " creating the AdminBean: "
                                     + e.getMessage()
                                     );
        } catch (Exception e) {
            e.printStackTrace();
            log.throwing("SyncBean", "loadAdminBean", e);
            throw new CreateException( "Error "
                                     + e.getClass().getName()
                                     + " creating the AdminBean: "
                                     + e.getMessage()
                                     );
        }
    }


    /**
     * When the session terminates this object is removed. This can happen
     * because the synchronization process has been completed or because the
     * session has expired. If the session completes its job, the synchro-
     * nization can be committed.
     *
     */
    public void ejbRemove() {
        syncAdapter.endSync();
    }

    public void ejbActivate() {
        log = Sync4jLogger.getLogger();
    }

    public void ejbPassivate() {
        log = null;
    }

    public void setSessionContext (SessionContext sessionContext)
    throws EJBException {
        this.sessionContext = sessionContext;
    }

    /** Processes an incoming XML message.
     *
     * @param requestURI the uri of the request
     * @param msg the SyncML request as an array of bytes
     * @param parameters SyncML request parameters
     * @param headers SyncML request headers
     *
     * @return the SyncML response as a <i>ISyncResponse</i> object
     *
     * @throws ServerException in case of a server error
     *
     */
    public SyncResponse processXMLMessage(final String requestURI,
                                          final byte[] msg       ,
                                          final Map    parameters,
                                          final Map    headers   )
    throws ServerException {
        return (SyncResponse) syncAdapter.processXMLMessage(requestURI,
                                                            msg,
                                                            parameters,
                                                            headers);
    }

    /** Processes an incoming WBXML message.
     *
     * @param requestURI the uri of the request
     * @param msg the SyncML request (WBXML encoded) as an array of bytes
     * @param parameters SyncML request parameters
     * @param headers SyncML request headers
     *
     * @return the SyncML response as a <i>ISyncResponse</i> object
     *
     * @throws ServerException in case of a server error
     *
     */
    public SyncResponse processWBXMLMessage(final String requestURI,
                                            final byte[] msg       ,
                                            final Map    parameters,
                                            final Map    headers   )
    throws ServerException {
        return (SyncResponse) syncAdapter.processWBXMLMessage(requestURI,
                                                              msg,
                                                              parameters,
                                                              headers);
    }

    /**
     * Used to process a status information as needed by the client object (i.e.
     * errors or success).
     *
     * @param statusCode the status code
     * @param info additional informational message
     *
     * @see com.funambol.framework.core.StatusCode for valid status codes.
     */
    public SyncResponse processStatusCode(int statusCode, String info){
        return syncAdapter.processStatusCode(statusCode, info);
    }
}
