/**
 * Copyright (C) 2003-2006 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the Honest Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Honest Public License for more details.
 *
 * You should have received a copy of the Honest Public License
 * along with this program; if not, write to Funambol,
 * 643 Bair Island Road, Suite 305 - Redwood City, CA 94063, USA
 */
package com.funambol.server.syncbean.session;

import java.io.File;

import java.util.ArrayList;
import java.util.Collection;
import java.util.logging.Logger;
import java.util.logging.Level;

import com.funambol.framework.tools.beans.BeanFactory;
import com.funambol.framework.tools.beans.BeanException;
import com.funambol.framework.logging.Sync4jLogger;
import com.funambol.framework.server.session.AbstractSessionManager;
import com.funambol.framework.server.session.SessionHandler;
import com.funambol.framework.server.session.SessionExpiredException;
import com.funambol.server.session.SyncSessionHandler;

/**
 * This implementation of a <i>AbstractSessionManager</i> stores and retrieves
 * sessions to/from the file system.
 * For each session is serialized under the directory pointed by <i>sessionsPath</i>
 * in a file called {session_id}.session . This is a very simple implementation
 * that allows many servers to access the same session repository. In fact, in
 * a real deployment, we can have SyncML servers replicated for reliability and
 * scalability purposes. In that case they must share the directory specified
 * by <i>sessionsPath</i>.
 * <p>
 * <b>This class is not thread safe. It is supposed that an instance is not
 * shared by multiple threads</b>.
 *
 * @version $Id: SessionManager.java,v 1.2 2006/08/07 21:13:18 nichele Exp $
 */
public class SessionManager extends AbstractSessionManager implements java.io.Serializable {

    // ------------------------------------------------------------ Private data

    /**
     * The logger
     */
    private static final Logger log = Sync4jLogger.getLogger();

    // -------------------------------------------------------------- Properties

    private String sessionsPath = null;

    /** Getter for property sessionsPath.
     * @return Value of property sessionsPath.
     */
    public String getSessionsPath() {
        return sessionsPath;
    }

    /** Setter for property sessionsPath.
     * @param sessionsPath New value of property sessionsPath.
     */
    public void setSessionsPath(String sessionsPath) {
        this.sessionsPath = sessionsPath;
    }

    // ---------------------------------------------------------- Public methods


    // --------------------------------------------------------- Private methods
    protected Collection getSessionBag() {
        ArrayList sessions = new ArrayList();

        //
        // Get a list of the content of the sessionsPath directory. Than for
        // each file deserialized it and add it to sessions.
        //
        File sessionsPathFile = new File(sessionsPath);

        if (!sessionsPathFile.exists()) {
            return sessions; // returns an empty collection
        }

        File[] files = sessionsPathFile.listFiles();

        SessionHandler handler = null;
        for (int i=0; (files != null) && (i<files.length); ++i) {
            try {
                handler = (SessionHandler)BeanFactory.getBeanInstance(files[i]);
                sessions.add(handler);
            } catch (BeanException e) {
                log.throwing(getClass().getName(), "getSessionBag", e.getCause());
            }
        }

        return sessions;
    }

    protected SessionHandler getSessionFromBag(String sessionId) throws SessionExpiredException {
        File sessionFile = new File(sessionsPath, sessionId + ".session");

        if (!sessionFile.exists()) {
            return null;
        }

        SessionHandler handler = null;
        try {
            handler = (SessionHandler)BeanFactory.getBeanInstance(sessionFile);
        } catch (BeanException e) {
            log.throwing(getClass().getName(), "getSessionFromBag", e.getCause());
        }

        return handler;
    }

    protected void putSessionInBag(String sessionId, SessionHandler handler) {
        File sessionFile = new File(sessionsPath, sessionId + ".session");

        try {
            BeanFactory.saveBeanInstance(handler, sessionFile);
        } catch (BeanException e) {
            log.severe("Error in serializing " + sessionFile);
            log.throwing(getClass().getName(), "putSessionInBag", e.getCause());
        }
    }

    protected void removeSessionFromBag(String sessionId) {
        File sessionFile = new File(sessionsPath, sessionId + ".session");

        if (sessionFile.exists()) {
            sessionFile.delete();
        }
    }

    protected SessionHandler createNewSession(String sessionId) {
        if (log.isLoggable(Level.FINEST)) {
            log.finest("Creating a new session with id " + sessionId);
        }

        SyncSessionHandler sessionHandler = new SyncSessionHandler();

        putSessionInBag(sessionId, sessionHandler);

        return sessionHandler;
    }
}

