/**
 * Copyright (C) 2003-2006 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the Honest Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Honest Public License for more details.
 *
 * You should have received a copy of the Honest Public License
 * along with this program; if not, write to Funambol,
 * 643 Bair Island Road, Suite 305 - Redwood City, CA 94063, USA
 */
package com.funambol.transport.http.server;

import java.io.*;
import java.net.NetworkInterface;
import java.util.Enumeration;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.Map;
import java.util.HashMap;
import java.util.Enumeration;
import java.util.logging.Logger;
import java.util.logging.Level;

import javax.servlet.*;
import javax.servlet.http.*;

import com.funambol.framework.core.*;
import com.funambol.framework.logging.Sync4jLogger;
import com.funambol.framework.server.*;
import com.funambol.framework.server.store.PersistentStore;
import com.funambol.framework.tools.Base64;
import com.funambol.framework.transport.http.SyncHolder;
import com.funambol.framework.protocol.ProtocolException;
import com.funambol.framework.security.Officer;
import com.funambol.framework.engine.SyncStrategy;
import com.funambol.framework.tools.IOTools;

import com.funambol.server.config.Configuration;
import com.funambol.transport.http.server.Constants;

import org.apache.commons.lang.StringUtils;

/**
 *  Receives the HTTP request and does session management.
 *
 *  <b>Session Management</b>
 *  <p>
 *  If the url contains a parameter {PARAM_SESSION_ID}={SESSION_ID}, the session
 *  id is used to lookup in the <i>handlerCache</i> if there is already a
 *  <i>SyncHolder</i> associated with it; the found handler is then used to process
 *  the incoming request. Otherwise, (either if the session parameter is not
 *  specified or no cached handler is found), a new session id is created and a
 *  new <i>SyncHolder</i> object is instantiated and stored in the cache.<br>
 *  The session id is created as followed:
 *  <p>
 *  <ol>
 *    <li> the first four bytes contains the IP address of the remote client
 *    <li> a dash ('-') is appended
 *    <li> the creation timestamp is appended
 *    <li> the resulting string is encoded base 64
 *  </ol>
 *  Note that no session expiration is handled by this class. That is delegated
 *  to the <i>SyncHandler</i> object.
 *
 *  @version $Id: ClusterableSync4jServlet.java,v 1.7 2006/08/07 21:13:18 nichele Exp $
 *
 */
public final class ClusterableSync4jServlet
extends HttpServlet
implements Constants {
    // ------------------------------------------------------------ Private data
    private String jndiAddress = DEFAULT_JNDI_ADDRESS;
    private RemoteHolderCache holderCache = null;

    private Logger log = Sync4jLogger.getLogger(LOG_NAME);

    //
    // To check if the engine was initialized
    //
    private boolean initialized = false;

    private static final int SIZE_INPUT_BUFFER = 4096;

    private static boolean logMessages    = false;
    private static String  dirlogMessages = null;

    //
    // The session timeout in seconds
    // (default value is 10 minutes...10*60 seconds)
    //
    private static int sessionTimeout = 10 * 60; // 10 minutes

    // ---------------------------------------------------------- Public methods

    public void init() throws ServletException {
        initialized = false;
        String value;

        if (log.isLoggable(Level.INFO)) {
            System.getProperties().list(System.out);
        }

        ServletConfig config = getServletConfig();

        value = config.getInitParameter(PARAM_CHANNEL_PROPERTIES);
        if (StringUtils.isEmpty(value)) {
            String msg = "The servlet configuration parameter "
                       + PARAM_CHANNEL_PROPERTIES
                       + " cannot be empty ("
                       + value
                       + ")"
                       ;
            log.severe(msg);
            throw new ServletException(msg);
        }

        value = config.getInitParameter(PARAM_SESSION_TIMEOUT);
        if (!StringUtils.isEmpty(value)) {
            try {
                sessionTimeout = Integer.parseInt(value);
            } catch (NumberFormatException e) {
                String msg = "The servlet configuration parameter "
                           + PARAM_SESSION_TIMEOUT
                           + " is not an integer number ("
                           + value
                           + ")"
                           ;
                log.severe(msg);
                throw new ServletException(msg);
            }
        }

        value = config.getInitParameter(PARAM_JNDI_ADDRESS);
        if (StringUtils.isEmpty(value)) {
            String msg = "Missing optional parameter "
                       + PARAM_JNDI_ADDRESS
                       + ", default value ("
                       + jndiAddress
                       + ") used."
                       ;
            log.warning(msg);
        } else {
            jndiAddress = value;
        }

        value = config.getInitParameter(PARAM_LOG_MESSAGES);
        if (value.equalsIgnoreCase("true")) {
            logMessages = true;

            dirlogMessages = config.getInitParameter(PARAM_DIRLOG_MESSAGES);
        }

        holderCache = new RemoteHolderCache(config);

        //
        // Now we are ready to bootstrap the Funambol components
        //
        bootstrap();

        //
        // Engine initialized!
        //
        initialized = true;
    }

    public void doPost(final HttpServletRequest  request ,
                       final HttpServletResponse response)
    throws ServletException, IOException {

        String requestURI = getRequestURI(request);

        if (log.isLoggable(Level.FINEST)) {
            showHeaders(request);
        }

        if (log.isLoggable(Level.INFO)) {
            StringBuffer sb = new StringBuffer("Handling incoming request: ");
            sb.append(requestURI);
            log.info(sb.toString());
            log.info("Requested sessionId: " + request.getRequestedSessionId());
        }

        if (request.getSession().isNew()) {
            request.getSession().setMaxInactiveInterval(sessionTimeout);
        }

        final String contentType = request.getContentType().split(";")[0];
        final int contentLength  = request.getContentLength();

        if (log.isLoggable(Level.FINEST)) {
            log.finest("contentType: " + contentType);
            log.finest("contentLength: " + contentLength);
        }

        if (contentLength < 1) {
            throw new Error("Content length < 1 (" + contentLength + ")");
        }

        Map    params     = getRequestParameters(request);
        Map    headers    = getRequestHeaders   (request);

        byte[] requestData = null;
        InputStream in = request.getInputStream();

        try{
            requestData = IOTools.readContent(in, SIZE_INPUT_BUFFER);
        } catch (Exception e) {
            handleError(request, response, "Error reading the request", e);
            return;
        } finally {

            try {
                if (in != null) {
                    in.close();
                }
            } catch (Exception ex) {
                handleError(request, response, ex.getClass().getName(), ex);
                return;
            }
            in = null;
        }

        //
        // If the session id is not specified in the URL, a new remote object
        // will be created. Otherwise the session id specifies which remote
        // object shall handles the request.
        //
        RemoteEJBSyncHolder holder = null;
        long requestTime = System.currentTimeMillis();
        String sessionId = null;
        try {
            holder    = createHolder(request);
            sessionId = holder.getSessionId();

            if (logMessages) {

                File fh = new File(dirlogMessages + File.separator +
                                   "RequestParameters_"   +
                                   requestTime            +
                                   "_"                    +
                                   sessionId              +
                                   ".txt"
                );

                BufferedWriter out = new BufferedWriter(new FileWriter(fh));
                out.write("**** Request Parameters: \r\n");
                java.util.Iterator itp = params.keySet().iterator();
                while(itp.hasNext()) {
                    String n = (String)itp.next();
                    out.write(n + ":" + (String)params.get(n) + "\r\n");
                }

                out.write("**** Request Headers   : \r\n");
                java.util.Iterator ith = headers.keySet().iterator();
                while(ith.hasNext()) {
                    String n = (String)ith.next();
                    out.write(n + ":" + (String)headers.get(n) + "\r\n");
                }
                out.close();

                File f = new File(dirlogMessages + File.separator +
                                  "RequestMsg_"  +
                                  requestTime    +
                                  "_"       +
                                  sessionId +
                                  ".bin"
                );
                FileOutputStream fos = new FileOutputStream(f);
                fos.write(requestData);
                fos.flush();
                fos.close();
            } //if logRequest
        } catch (Exception e) {
            handleError(request, response, "Error creating SyncBean", e);
            return;
        }

        SyncResponse resp = null;
        try {
            if (com.funambol.framework.core.Constants.MIMETYPE_SYNCMLDS_WBXML.equals(contentType)) {
                resp = holder.processWBXMLMessage(requestURI, requestData, params, headers);
            } else if (com.funambol.framework.core.Constants.MIMETYPE_SYNCMLDS_XML.equals(contentType)) {
                resp = holder.processXMLMessage(requestURI, requestData, params, headers);
            } else {
                throw new ProtocolException( "Mime type "
                                           + contentType
                                           + " nor supported or unknown" );
            }
        } catch (Exception e) {
            Throwable cause = e.getCause();

            if (  (cause != null)
               && (  (cause instanceof ProtocolException)
                  || (cause instanceof Sync4jException  )
                  )
               ) {
                handleError(request, response, "Protocol error", cause);
                return;
            } else {
                throw new ServletException(e);
            }
        } finally {
            //
            // Now we can store the handler into the distributed cache. Note
            // that when the holder is put in the cache a serialized copy is
            // actually stored, not the simple local reference.
            //
            if (holder.isNew()) {
                holderCache.put(holder);
            }
        }

        OutputStream out = null;
        try {
            out = response.getOutputStream();
            response.setContentType(contentType);

            byte[] buf = resp.getMessage();
            if (log.isLoggable(Level.FINEST)) {
                log.finest("Outgoing message length: " + buf.length);
            }

            if (logMessages) {

                File f = new File(dirlogMessages + File.separator +
                                  "ResponseMsg_" +
                                  requestTime    +
                                  "_"            +
                                  sessionId      +
                                  ".bin"
                         );
                FileOutputStream fos = new FileOutputStream(f);
                fos.write(buf);
                fos.flush();
                fos.close();
            }

            response.setContentLength(buf.length);
            out.write(buf);
            out.flush();

        } finally {
            if (log.isLoggable(Level.FINEST)) {
                log.finest("Finally");
            }
            if (out != null) {
                out.close();
            }
        }

        //
        // If the message completed the SyncML communication, the holder
        // must be closed and discarded.
        //
        if (resp.isCompleted()) {
            releaseHolder(holder);
        }
    }

    public void doGet(final HttpServletRequest  httpRequest,
                      final HttpServletResponse httpResponse)
    throws ServletException, IOException {

        String pathInfo   = httpRequest.getPathInfo();

        if (log.isLoggable(Level.FINEST)) {
            log.finest("pathInfo: " + pathInfo);
        }

        if (PATH_INFO_STATUS.equals(pathInfo)) {
            if (initialized) {
                //
                //The request sent by the client was successful
                //
                httpResponse.setStatus(httpResponse.SC_OK);
                if (log.isLoggable(Level.FINEST)) {
                    log.finest("Request succeeded normally");
                }
            } else {
                //
                //The request was unsuccessful to the server being down or overloaded.
                //
                httpResponse.setStatus(httpResponse.SC_SERVICE_UNAVAILABLE);
                if (log.isLoggable(Level.FINEST)) {
                    log.finest("The HTTP server is unable to handle the request");
                }
            }
        }
    }

    // ------------------------------------------------------- Protected methods

    // --------------------------------------------------------- Private methods
    /**
     * Factory method for <i>SyncHolder</i> objects. If the session id is
     * passed as a CGI parameter, it is transformed in a EJB handle. If the
     * handle is not valid, it is considered expired. If the session id is not
     * specified, a new EJB is created
     *
     * @param request the associated HTTP request object
     *
     * @return a new <i>SyncHolder</i>
     *
     */
    private RemoteEJBSyncHolder createHolder(HttpServletRequest request)
    throws Exception {
        String sessionId = request.getParameter(PARAM_SESSION_ID);

        RemoteEJBSyncHolder holder = null;

        if (log.isLoggable(Level.FINEST)) {
            log.finest("holderCache: " + holderCache);
        }

        if (!StringUtils.isEmpty(sessionId)) {

            if (log.isLoggable(Level.FINEST)) {
                log.finest("Searching holder with sessionId: " + sessionId);
            }

            holder = (RemoteEJBSyncHolder)holderCache.get(sessionId);
        }

        if (holder == null) {

            if (log.isLoggable(Level.FINEST)) {
                log.finest("Holder not found. A new holder will be created");
            }

            holder = new RemoteEJBSyncHolder(true);

            holder.setSessionId(createSessionId(request));
            holder.setJndiAddress(jndiAddress);
        } else {
            if (log.isLoggable(Level.FINEST)) {
                log.finest("Holder found");
            }
        }

        return holder;
    }

    /**
     * Closes the given <i>SyncHolder</i> and removes it from the cache.
     *
     * @param holder the holder to releases
     */
    private void releaseHolder(SyncHolder holder) {
        try {
            holder.close();
        } catch (Exception e) {
            log.severe(e.getMessage());
            log.throwing(getClass().getName(), "releaseHolder", e.getCause());
        }
        holderCache.remove(holder.getSessionId());
    }

    /**
     * Handles errors conditions returning an appropriate content to the client.
     *
     * @param request the request object
     * @param response the response object
     * @msg   a desctiptive message
     * @t     a throwable object
     *
     */
    private void handleError(final HttpServletRequest   request,
                             final HttpServletResponse response,
                                   String                   msg,
                             final Throwable                  t) {

        if (msg == null) {
            msg = "";
        }

        if (t == null) {
           log.severe(msg);
        } else {
            log.severe(msg);
            log.throwing(getClass().getName(), "handleError", t);
        }
        try {
            response.sendError(response.SC_BAD_REQUEST, msg);
        } catch (IOException e) {
            log.severe(e.getMessage());
            log.throwing(getClass().getName(), "handleError", e);
        }
    }

    /**
     * Creates the session id (see the class description for details).
     *
     * @param request the HTTP request object
     *
     * @return a newly created session id
     */
    private String createSessionId(HttpServletRequest request) {
        String clientIP  = request.getRemoteAddr()   ;
        long   timestamp = System.currentTimeMillis();

        StringBuffer sb = new StringBuffer();

        byte[] ip = null;

        try {
            ip = InetAddress.getByName(clientIP).getAddress();
        } catch (UnknownHostException e) {
            ip = new byte[] {0, 0, 0, 0};
        }
        //append to the session id the bytes in the ip
        for (int i = 0; i < ip.length; i++) {
            sb.append(ip[i]);
            if (i < ip.length - 1) {
                sb.append(".");
            }
        }
        sb.append('-').append(timestamp);

        String stringSessionId = new String(Base64.encode(sb.toString().getBytes()));

        //
        // strips out tailing '='
        //
        int i = stringSessionId.lastIndexOf('=');

        return ((i>0) ? stringSessionId.substring(0, i-1) : stringSessionId);
    }

    /**
     * Extracts the request parameters and returns them in a <i>Map</i>
     *
     * @param request the request
     */
    private Map getRequestParameters(HttpServletRequest request) {
        Map params = new HashMap();

        String paramName = null;
        for (Enumeration e=request.getParameterNames(); e.hasMoreElements(); ) {
            paramName = (String)e.nextElement();

            params.put(paramName, request.getParameter(paramName));
        }

        return params;
    }

    /**
     * Extracts the request headers and returns them in a <i>Map</i>
     *
     * @param request the request
     */
    private Map getRequestHeaders(HttpServletRequest request) {
        Map headers = new HashMap();

        String headerName = null;
        for (Enumeration e=request.getHeaderNames(); e.hasMoreElements(); ) {
            headerName = (String)e.nextElement();

            headers.put(headerName, request.getHeader(headerName));
        }

        return headers;
    }

    /**
     * Bootstraps the Funambol server components and logs the server startup
     * event on both the console and the logging system.
     */
    private void bootstrap() throws ServletException {
        try {
            Configuration c = null;

            c = Configuration.getConfiguration();

            String msg1 = "================================================================================";
            String sp   = "\n\n";
            String msg2 = "SyncServer v. "
                        + c.getServerConfig().getServerInfo().getSwV()
                        + " engine started.\nConfiguration object found."
                        ;

            //
            // NOTE: this code uses the stdout by choice, so that it will always
            // be displayed, regardless how logging is configured.
            //
            System.out.println(msg1);
            System.out.println(sp);
            System.out.println(msg2);
            System.out.println(sp);
            System.getProperties().list(System.out);
            System.out.println(sp);
            System.out.println(msg1);

            if (log.isLoggable(Level.INFO)) {
                log.info(msg1+sp);
                log.info(msg2+sp);
                log.info(System.getProperties().toString()+sp);
                log.info(msg1);
            }

            String store       = c.getServerConfig().getEngineConfiguration().getStore();
            PersistentStore ps = c.getStore();
            if (ps == null) {
                throw new ServletException("Fatal error creating the PersistentStore object");
            }

            String officer = c.getServerConfig().getEngineConfiguration().getOfficer();
            Officer off    = c.getOfficer();
            if (off == null) {
                throw new ServletException("Fatal error creating the Officer object");
            }

            String strategy  = c.getServerConfig().getEngineConfiguration().getStrategy();
            SyncStrategy str = c.getStrategy();
            if (str == null) {
                throw new ServletException("Fatal error creating the SyncStrategy object");
            }

            if (log.isLoggable(Level.INFO)) {
                log.info("Engine configuration:");
                log.info(" - store: "
                          + store
                          + " (" + ps + ")"
                );
                log.info(" - officer: "
                          + officer
                          + " (" + off + ")"
                );
                log.info(" - strategy: "
                          + strategy
                          + " (" + str + ")"
                );

                log.info("Default encoding: "
                         + (new OutputStreamWriter(new ByteArrayOutputStream()))
                         .getEncoding()
                );

                String serverURI =
                    c.getServerConfig().getEngineConfiguration().getServerURI();

                if (StringUtils.isEmpty(serverURI)) {
                    log.info("Server URI not set.");
                } else {
                    log.info("Server URI set: " + serverURI);
                }
            }
        } catch (Exception e) {
            throw new ServletException(e.getMessage());
        }
    }

    /**
     * Log the headers of the request
     * @param request HttpServletRequest
     */
    private void showHeaders(HttpServletRequest request) {
        Enumeration enumHeaders = request.getHeaderNames();
        String headerName = null;
        log.finest("Http header: ");
        while (enumHeaders.hasMoreElements()) {
            headerName = (String)enumHeaders.nextElement();
            log.finest(headerName + ": " + request.getHeader(headerName));
        }
    }

    /**
     * Returns the request uri getting the request url and the query string
     * @param request HttpServletRequest
     * @return String
     */
    private String getRequestURI(HttpServletRequest request) {
        StringBuffer sb = request.getRequestURL();
        String queryString = request.getQueryString();
        if (queryString != null) {
            sb.append("?").append(queryString);
        }
        return sb.toString();
    }
}
