/**
 * Copyright (C) 2003-2006 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the Honest Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Honest Public License for more details.
 *
 * You should have received a copy of the Honest Public License
 * along with this program; if not, write to Funambol,
 * 643 Bair Island Road, Suite 305 - Redwood City, CA 94063, USA
 */
package com.funambol.transport.http.server;

import java.util.Map;
import java.util.logging.Logger;

import com.funambol.framework.config.ConfigurationException;
import com.funambol.framework.core.Sync4jException;
import com.funambol.framework.logging.Sync4jLogger;
import com.funambol.framework.logging.Sync4jLoggerName;
import com.funambol.framework.server.SyncResponse;
import com.funambol.framework.server.error.ServerException;
import com.funambol.framework.transport.http.SyncHolder;

import com.funambol.server.config.Configuration;
import com.funambol.server.engine.SyncAdapter;

/**
 * Implements a <i>SyncHolder</i> that instance a SyncAdapter to handle a
 * synchronization request.
 *
 * @version $Id: LocalSyncHolder.java,v 1.4 2006/08/07 21:13:18 nichele Exp $
 */
public class LocalSyncHolder implements SyncHolder {

    // --------------------------------------------------------------- Constants

    // ------------------------------------------------------------ Private data

    private transient Logger log = Sync4jLogger.getLogger(Sync4jLoggerName.HANDLER);

    private SyncAdapter syncAdapter       = null;
    private long        creationTimestamp       ;
    private String      sessionId               ;

    // ------------------------------------------------------------ Constructors
    public LocalSyncHolder() throws ServerException {
        try {
            syncAdapter = new SyncAdapter(loadConfiguration());
        } catch (ConfigurationException e) {

        }
        creationTimestamp = System.currentTimeMillis();
    }

    // ---------------------------------------------------------- Public methods

    /**
     * Processes an incoming XML message.
     *
     * @param requestURI the uri of the request
     * @param requestData the SyncML request as an array of bytes
     * @param parameters SyncML request parameters
     * @param headers SyncML request headers
     *
     * @return the SyncML response as a <i>ISyncResponse</i> object
     *
     * @throws ServerException in case of a server error
     *
     */
    public SyncResponse processXMLMessage(final String requestURI,
                                          final byte[] requestData,
                                          final Map    parameters ,
                                          final Map    headers    )
    throws ServerException {
        return syncAdapter.processXMLMessage(requestURI, requestData, parameters, headers);
    }

    /** Processes an incoming WBXML message.
     *
     * @param requestURI the uri of the request
     * @param requestData the SyncML request as an array of bytes
     * @param parameters SyncML request parameters
     * @param headers SyncML request headers
     *
     * @return the SyncML response as a <i>ISyncResponse</i> object
     *
     * @throws ServerException in case of a server error
     *
     */
    public SyncResponse processWBXMLMessage(final String requestURI,
                                            final byte[] requestData,
                                            final Map    parameters ,
                                            final Map    headers    )

    throws ServerException {
        return syncAdapter.processWBXMLMessage(requestURI, requestData, parameters, headers);
    }

    public void setSessionId(String sessionId) throws Sync4jException {
        this.sessionId = sessionId;
        syncAdapter.beginSync(sessionId);
    }

    public String getSessionId() {
        return this.sessionId;
    }

    /** Called when the SyncHolder is not required any more. It gives the holder
     * an opportunity to release and clean up resources.
     *
     * @throws java.lang.Exception in case of error. The real exception is stored
     * in the cause.
     *
     */
    public void close() throws Exception {
        syncAdapter.endSync();
    }

    /**
     * Returns the creation timestamp (in milliseconds since midnight, January
     * 1, 1970 UTC).
     *
     * @see com.funambol.framework.transport.http.SyncHolder
     */
    public long getCreationTimestamp() {
        return creationTimestamp;
    }

    /**
     * Loads the configuration for the server.
     *
     * @throws ConfigurationException in case of errors.
     */
    private Configuration loadConfiguration()
    throws ConfigurationException {

        Configuration config = Configuration.getConfiguration();

        //
        // If config is null, being Configuration a singleton, only an exception
        // could have been happend.
        //
        if (config == null) {
            throw new ConfigurationException("Error loading configuration (see the log for details)!");
        }

        return config;
    }
}
