/**
 * Copyright (C) 2003-2006 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the Honest Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Honest Public License for more details.
 *
 * You should have received a copy of the Honest Public License
 * along with this program; if not, write to Funambol,
 * 643 Bair Island Road, Suite 305 - Redwood City, CA 94063, USA
 */
package com.funambol.transport.http.server;

import java.io.*;
import java.net.NetworkInterface;
import java.util.Enumeration;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.Map;
import java.util.HashMap;
import java.util.Enumeration;
import java.util.logging.Logger;
import java.util.logging.Level;

import javax.servlet.*;
import javax.servlet.http.*;

import com.funambol.framework.core.*;
import com.funambol.framework.logging.Sync4jLogger;
import com.funambol.framework.server.*;
import com.funambol.framework.server.store.PersistentStore;

import com.funambol.framework.transport.http.SyncHolder;
import com.funambol.framework.transport.http.SyncHolderCache;
import com.funambol.framework.protocol.ProtocolException;
import com.funambol.framework.security.Officer;
import com.funambol.framework.engine.SyncStrategy;
import com.funambol.framework.tools.IOTools;

import com.funambol.server.config.Configuration;

import org.apache.commons.lang.StringUtils;

/**
 * Receives the HTTP request and does session management.
 *
 * <b>Session Management</b>
 * <p>
 * If the url contains a parameter {PARAM_SESSION_ID}={SESSION_ID}, the session
 * id is used to lookup in the <i>handlerCache</i> if there is already a
 * <i>SyncHolder</i> associated with the given session id it is used to process
 * the incoming request. Otherwise (either if the session parameter is not
 * specified or not cached handler is found), a new session id is created and a
 * new <i>SyncHolder</i> object is instantiated and stored in the cache.<br/>
 * The sessionId is created by the <code>SessionTools</code> using, if there are,
 * the request attribute
 * <ui>
 * <li><code>funambol.server_port</code></li>
 * <li><code>funambol.server_ip</code></li>
 * </ui>
 * <br/>
 *
 * Note that no session expiration is handled by this class. That is delegated
 * to the <i>SyncHolderCache</i> object.
 *
 * @version $Id: Sync4jServlet.java,v 1.14 2006/08/10 13:48:20 nichele Exp $
 */
public final class Sync4jServlet extends HttpServlet implements Constants {

    // ------------------------------------------------------------ Private data

    private Logger log = Sync4jLogger.getLogger(LOG_NAME);

    private static String          syncHolderClass = null;
    private static SyncHolderCache holderCache     = null;

    //
    // To check if the engine was initialized
    //
    private boolean initialized = false;

    private static final int SIZE_INPUT_BUFFER = 4096;

    private static boolean logMessages    = false;
    private static String  dirlogMessages = null;

    //
    // The session timeout in seconds
    // (default value is 10 minutes...10*60 seconds)
    //
    private static int sessionTimeout = 10 * 60; // 10 minutes

    // ---------------------------------------------------------- Public methods

    public void init() throws ServletException {
        initialized = false;
        String value;

        if (log.isLoggable(Level.INFO)) {
            System.getProperties().list(System.out);
        }

        ServletConfig config = getServletConfig();

        value = config.getInitParameter(PARAM_SYNCHOLDER_CLASS);

        if (StringUtils.isEmpty(value)) {
            String msg = "The servlet configuration parameter "
            + PARAM_SYNCHOLDER_CLASS
            + " cannot be empty ("
            + value
            + ")"
            ;
            log(msg);
            throw new ServletException(msg);
        }
        syncHolderClass = value;

        value = config.getInitParameter(PARAM_SESSION_TIMEOUT);
        if (!StringUtils.isEmpty(value)) {
            try {
                sessionTimeout = Integer.parseInt(value);
            } catch (NumberFormatException e) {
                String msg = "The servlet configuration parameter "
                + PARAM_SESSION_TIMEOUT
                + " is not an integer number ("
                + value
                + ")"
                ;
                log.severe(msg);
                throw new ServletException(msg);
            }
        }

        value = config.getInitParameter(PARAM_LOG_MESSAGES);
        if (value.equalsIgnoreCase("true")) {
            logMessages = true;
            dirlogMessages = config.getInitParameter(PARAM_DIRLOG_MESSAGES);
        }

        //
        // sessionTimeout is in seconds
        //
        holderCache = new SyncHolderCache(sessionTimeout * 1000);

        //
        // Now we are ready to bootstrap the Funambol components
        //
        bootstrap();

        //
        // Engine initialized!
        //
        initialized = true;
    }


    public void doPost(final HttpServletRequest  httpRequest ,
                       final HttpServletResponse httpResponse)
    throws ServletException, IOException {

        if (log.isLoggable(Level.FINEST)) {
            showHeaders(httpRequest);
        }

        String requestURL = getRequestURL(httpRequest);

        if (log.isLoggable(Level.INFO)) {
            StringBuffer sb = new StringBuffer("Handling incoming request: ");
            sb.append(requestURL);
            log.info(sb.toString());
            log.info("Requested sessionId: " + httpRequest.getRequestedSessionId());
        }

        if (httpRequest.getSession().isNew()) {
            httpRequest.getSession().setMaxInactiveInterval(sessionTimeout);
        }

        //
        // Setting the header 'Set-Cookie' in order to avoid the session tracking
        // using cookies.
        //
        // The web container adds a cookie JSESSIONID in order to
        // track the session, and to do that, it adds (transparently) in the response
        // header:
        // Set-Cookie: JSESSIONID=xxxxxxxxxx
        // In order not to use the cookie, the header "Set-Cookie" is set to an empty value
        // In this way, the session tracking is based on the jsessionid parameter
        // specified in the url (url rewriting).
        // The cookie is dangerous because a client could use:
        // http://xxxxxx:yyy/funambol/ds
        // but with a jsessionid specified as cookie. In this way, the server
        // search a previous session with the same id. And if a previous session
        // was interrupted and not expired, the server reuses this one and this
        // can cause an exception because the client sends the msg 1 and maybe
        // the previous session was in the mapping state.
        //
        httpResponse.setHeader("Set-Cookie", "");

        final String contentType = httpRequest.getContentType().split(";")[0];
        final int contentLength  = httpRequest.getContentLength();

        if (log.isLoggable(Level.FINEST)) {
            log.finest("contentType: " + contentType);
            log.finest("contentLength: " + contentLength);
        }

        Map    params     = getRequestParameters(httpRequest);
        Map    headers    = getRequestHeaders   (httpRequest);

        byte[] requestData = null;
        InputStream in = httpRequest.getInputStream();

        try{
            requestData = IOTools.readContent(in, SIZE_INPUT_BUFFER);
        } catch (Exception e) {
            handleError(httpRequest, httpResponse, "Error reading the request", e);
            return;
        } finally {

            try {
                if (in != null) {
                    in.close();
                }
            } catch (Exception ex) {
                handleError(httpRequest, httpResponse, ex.getClass().getName(), ex);
                return;
            }
            in = null;
        }

        //
        // If the session id is not specified in the URL, a new remote object
        // will be created. Otherwise the session id specifies which remote
        // object shall handles the request.
        //
        SyncHolder holder = null;
        long requestTime  = System.currentTimeMillis();
        String sessionId  = null;

        try {
            holder    = createHolder(httpRequest);
            sessionId = holder.getSessionId();

            if (logMessages) {

                File fh = new File(dirlogMessages + File.separator +
                                   "RequestParameters_"  +
                                   requestTime           +
                                   "_"                   +
                                   sessionId             +
                                   ".txt"
                );

                BufferedWriter out = new BufferedWriter(new FileWriter(fh));
                out.write("**** Request Parameters: \r\n");
                java.util.Iterator itp = params.keySet().iterator();
                while(itp.hasNext()) {
                    String n = (String)itp.next();
                    out.write(n + ":" + (String)params.get(n) + "\r\n");
                }

                out.write("**** Request Headers   : \r\n");
                java.util.Iterator ith = headers.keySet().iterator();
                while(ith.hasNext()) {
                    String n = (String)ith.next();
                    out.write(n + ":" + (String)headers.get(n) + "\r\n");
                }
                out.close();

                File f = new File(dirlogMessages + File.separator +
                                  "RequestMsg_"  +
                                  requestTime    +
                                  "_"            +
                                  sessionId      +
                                  ".bin"
                );
                FileOutputStream fos = new FileOutputStream(f);
                fos.write(requestData);
                fos.flush();
                fos.close();
            } //if logRequest
        } catch (Exception e) {
            handleError(httpRequest, httpResponse, "Error creating SyncBean", e);
            return;
        }

        SyncResponse resp = null;
        try {
            if (com.funambol.framework.core.Constants.MIMETYPE_SYNCMLDS_WBXML.equals(contentType)) {
                resp = holder.processWBXMLMessage(requestURL, requestData, params, headers);
            } else if (com.funambol.framework.core.Constants.MIMETYPE_SYNCMLDS_XML.equals(contentType)) {
                resp = holder.processXMLMessage(requestURL, requestData, params, headers);
            } else {
                throw new ProtocolException( "Mime type "
                                           + contentType
                                           + " not supported or unknown" );
            }
        } catch (Exception e) {
            Throwable cause = e.getCause();

            if ( (cause != null) &&
                 ((cause instanceof ProtocolException) || (cause instanceof Sync4jException)) ) {

                handleError(httpRequest, httpResponse, "Protocol error", cause);
                return;
            } else {
                throw new ServletException(e);
            }
        }

        OutputStream out = null;
        try {
            out = httpResponse.getOutputStream();
            httpResponse.setContentType(contentType);

            byte[] response = resp.getMessage();

            if (log.isLoggable(Level.FINEST)) {
                log.finest("Outgoing message length: " + response.length);
            }

            if (logMessages) {

                File f = new File(dirlogMessages + File.separator +
                                  "ResponseMsg_" +
                                  requestTime    +
                                  "_"            +
                                  sessionId      +
                                  ".bin"
                         );
                FileOutputStream fos = new FileOutputStream(f);
                fos.write(response);
                fos.flush();
                fos.close();
            }

            httpResponse.setContentLength(response.length);
            out.write(response);
            out.flush();

        } finally {
            if (out != null) {
                out.close();
            }
            if (log.isLoggable(Level.FINEST)) {
                log.finest("Finally");
            }
        }

        //
        // If the message completed the SyncML communication, the holder
        // must be closed and discarded.
        //
        if (resp.isCompleted()) {
            releaseHolder(holder);
        }
    }

    public void doGet(final HttpServletRequest  httpRequest ,
                      final HttpServletResponse httpResponse)
    throws ServletException, IOException {

        String pathInfo   = httpRequest.getPathInfo();

        if (PATH_INFO_STATUS.equals(pathInfo)) {
            if (initialized) {
                //
                //The request sent by the client was successful
                //
                httpResponse.setStatus(httpResponse.SC_OK);
                if (log.isLoggable(Level.FINEST)) {
                    log.finest("Request succeeded normally");
                }
            } else {
                //
                //The request was unsuccessful to the server being down or overloaded.
                //
                httpResponse.setStatus(httpResponse.SC_SERVICE_UNAVAILABLE);
                if (log.isLoggable(Level.FINEST)) {
                    log.finest("The HTTP server is unable to handle the request");
                }
            }
        }
    }

    // ------------------------------------------------------- Protected methods

    // --------------------------------------------------------- Private methods
    /**
     * Factory method for <i>SyncHolder</i> objects. If the session id is
     * passed as a CGI parameter, it is transformed in a EJB handle. If the
     * handle is not valid, it is considered expired. If the session id is not
     * specified, a new EJB is created
     *
     * @param request the associated HTTP request object
     *
     * @return a new <i>SyncHolder</i>
     *
     */
    private SyncHolder createHolder(HttpServletRequest request)
    throws Exception {

        String sessionId = null;

        sessionId = getSessionId(request);
        if (sessionId == null) {
            sessionId = request.getSession().getId();
        }

        SyncHolder holder = null;

        if (log.isLoggable(Level.FINEST)) {
            log.finest("cache: " + holderCache);
        }

        if (!StringUtils.isEmpty(sessionId)) {
            if (log.isLoggable(Level.FINEST)) {
                log.finest("Searching holder with sessionId: " + sessionId);
            }
            holder = (SyncHolder)holderCache.get(sessionId);
        }

        if (holder == null) {

            if (log.isLoggable(Level.FINEST)) {
                log.finest("Holder not found. A new holder will be created");
            }

            holder = (SyncHolder)getClass().forName(syncHolderClass).newInstance();

            holder.setSessionId(sessionId);
            holderCache.put(holder);
        } else {
            if (log.isLoggable(Level.FINEST)) {
                log.finest("Holder found");
            }

        }

        return holder;
    }

    /**
     * Closes the given <i>SyncHolder</i> and removes it from the cache.
     *
     * @param holder the holder to releases
     */
    private void releaseHolder(SyncHolder holder) {
        try {
            holder.close();
        } catch (Exception e) {
            log.severe(e.getMessage());
            log.throwing(getClass().getName(), "releaseHolder", e.getCause());
        }
        holderCache.remove(holder.getSessionId());
    }

    /**
     * Handles errors conditions returning an appropriate content to the client.
     *
     * @param request the request object
     * @param response the response object
     * @msg   a desctiptive message
     * @t     a throwable object
     *
     */
    private void handleError(final HttpServletRequest  request ,
                             final HttpServletResponse response,
                             String                    msg     ,
                             final Throwable           t       ) {

        if (msg == null) {
            msg = "";
        }

        if (t == null) {
            log.severe(msg);
        } else {
            log.severe(msg);
            log.throwing(getClass().getName(), "unknown", t);
        }
        try {
            response.sendError(response.SC_BAD_REQUEST, msg);
        } catch (IOException e) {
            log.severe(e.getMessage());
            log.throwing(getClass().getName(), "unknown", t);
        }
    }


    /**
     * Extracts the request parameters and returns them in a <i>Map</i>
     *
     * @param request the request
     */
    private Map getRequestParameters(HttpServletRequest request) {
        Map params = new HashMap();

        String paramName = null;
        for (Enumeration e=request.getParameterNames(); e.hasMoreElements(); ) {
            paramName = (String)e.nextElement();

            params.put(paramName, request.getParameter(paramName));
        }

        return params;
    }

    /**
     * Extracts the request headers and returns them in a <i>Map</i>
     *
     * @param request the request
     */
    private Map getRequestHeaders(HttpServletRequest request) {
        Map headers = new HashMap();

        String headerName = null;
        for (Enumeration e=request.getHeaderNames(); e.hasMoreElements(); ) {
            headerName = (String)e.nextElement();

            headers.put(headerName.toLowerCase(), request.getHeader(headerName));
        }

        return headers;
    }

    /**
     * Bootstraps the Funambol server components and logs the server startup
     * event on both the console and the logging system.
     */
    private void bootstrap() throws ServletException {
        try {
            Configuration c = null;

            c = Configuration.getConfiguration();

            String msg1 = "================================================================================";
            String sp   = "\n\n";
            String msg2 = "Funambol Data Synchronization Server v. "
                    + c.getServerConfig().getServerInfo().getSwV()
                    + " engine started.\nConfiguration object found."
                    ;

            //
            // NOTE: this code uses the stdout by choice, so that it will always
            // be displayed, regardless how logging is configured.
            //
            System.out.println(msg1);
            System.out.println(sp);
            System.out.println(msg2);
            System.out.println(sp);
            System.getProperties().list(System.out);
            System.out.println(sp);
            System.out.println(msg1);

            if (log.isLoggable(Level.INFO)) {
                log.info(msg1+sp);
                log.info(msg2+sp);
                log.info(System.getProperties().toString()+sp);
                log.info(msg1);
            }

            String store       = c.getServerConfig().getEngineConfiguration().getStore();
            PersistentStore ps = c.getStore();
            if (ps == null) {
                throw new ServletException("Fatal error creating the PersistentStore object");
            }

            String officer = c.getServerConfig().getEngineConfiguration().getOfficer();
            Officer off    = c.getOfficer();
            if (off == null) {
                throw new ServletException("Fatal error creating the Officer object");
            }

            String strategy  = c.getServerConfig().getEngineConfiguration().getStrategy();
            SyncStrategy str = c.getStrategy();
            if (str == null) {
                throw new ServletException("Fatal error creating the SyncStrategy object");
            }

            if (log.isLoggable(Level.INFO)) {
                log.info("Engine configuration:");

                log.info(" - store: "
                         + store
                         + " (" + ps + ")"
                        );

                log.info(" - officer: "
                         + officer
                         + " (" + off + ")"
                        );

                log.info(" - strategy: "
                         + strategy
                         + " (" + str + ")"
                        );

                log.info("Default encoding: "
                         + (new OutputStreamWriter(new ByteArrayOutputStream()))
                         .getEncoding()
                        );

                String serverURI =
                        c.getServerConfig().getEngineConfiguration().getServerURI();

                if (StringUtils.isEmpty(serverURI)) {
                    log.info("Server URI not set.");
                } else {
                    log.info("Server URI set: " + serverURI);
                }
            }
        } catch(Exception e) {
            throw new ServletException(e.getMessage());
        }
    }

    /**
     * Log the headers of the request
     * @param request HttpServletRequest
     */
    private void showHeaders(HttpServletRequest request) {
        Enumeration enumHeaders = request.getHeaderNames();
        String headerName = null;
        log.finest("Http header: ");
        while (enumHeaders.hasMoreElements()) {
            headerName = (String)enumHeaders.nextElement();
            log.finest(headerName + ": " + request.getHeader(headerName));
        }
    }

    /**
     * Returns the sessionId of the request. <br/>
     * The sessionId is extracted from the pathInfo or, if there isn't there,
     * we use the id of the <code>HttpSession</code>
     * @param request HttpServletRequest
     * @return String
     */
    private String getSessionId(HttpServletRequest request) {
        String pathInfo = request.getPathInfo();
        String sessionId = null;

        if (StringUtils.isNotEmpty(pathInfo)) {
            int index = pathInfo.indexOf(";jsessionid=");
            if (index != -1) {
                sessionId = pathInfo.substring(index + 12);
            }
        }
        if (sessionId == null) {
            sessionId = request.getSession().getId();
        }
        return sessionId;
    }

    /**
     * Returns the request uri getting the request url and the query string
     * @param request HttpServletRequest
     * @return String
     */
    private String getRequestURL(HttpServletRequest request) {
        StringBuffer sb = request.getRequestURL();
        String queryString = request.getQueryString();
        if (queryString != null) {
            sb.append("?").append(queryString);
        }
        return sb.toString();
    }

}
